import io
import json
import os
import threading
import urllib.parse
import webbrowser
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer

import numpy as np

# Use Agg backend (headless-safe)
import matplotlib
matplotlib.use("Agg")
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas


_HTML = """<!doctype html>
<html>
<head>
  <meta charset="utf-8">
  <title>Inference – GT vs Prediction</title>
  <style>
    body { font-family: system-ui, -apple-system, Segoe UI, Roboto, sans-serif; margin: 16px; background:#111; color:#eee; }
    .controls { display:flex; gap:12px; align-items:center; flex-wrap:wrap; margin-bottom:12px; }
    button { padding:6px 10px; border:none; border-radius:6px; cursor:pointer; background:#2E86AB; color:white; }
    button.secondary { background:#F18F01; }
    button.danger { background:#A23B72; }
    label { font-size:12px; color:#ccc; }
    input[type=number] { width:4rem; margin-left:4px; }
    .panel { display:flex; flex-direction:column; align-items:flex-start; gap:8px; }
    img { background:black; max-width:95vw; height:auto; border-radius:8px; }
    .meta { font-size:12px; color:#ccc; }
    .indices { font-family: ui-monospace, SFMono-Regular, Menlo, monospace; font-size:12px; color:#ddd; }
  </style>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <meta http-equiv="Cache-Control" content="no-store" />
  <meta http-equiv="Pragma" content="no-cache" />
  <meta http-equiv="Expires" content="0" />
</head>
<body>
  <div class="controls">
    <button id="shuffle" class="danger">Shuffle Slices</button>
    <button id="reset" class="secondary">Reset</button>
    <button id="savefig">Save Figure</button>
    <label>Z <input id="zidx" type="number" value="0" min="0" step="1"></label>
    <label>Y <input id="yidx" type="number" value="0" min="0" step="1"></label>
    <label>X <input id="xidx" type="number" value="0" min="0" step="1"></label>
    <span id="savehint" class="meta"></span>
  </div>
  <div class="panel">
    <img id="grid" src="" alt="grid">
    <div id="meta" class="meta"></div>
    <div class="indices" id="indices"></div>
  </div>

<script>
let meta = null;
let z = 0, y = 0, x = 0;

function qs(sel){ return document.querySelector(sel); }
function clamp(v, lo, hi){ v = parseInt(v||0,10); return Math.max(lo, Math.min(hi, v)); }

async function loadMeta() {
  const r = await fetch('meta');
  meta = await r.json();
  z = meta.mid.z; y = meta.mid.y; x = meta.mid.x;
  qs('#zidx').max = meta.dims[0]-1; qs('#zidx').value = z;
  qs('#yidx').max = meta.dims[1]-1; qs('#yidx').value = y;
  qs('#xidx').max = meta.dims[2]-1; qs('#xidx').value = x;
  qs('#meta').innerText = `Dims: ${meta.dims.join(' x ')} | Classes: ${meta.classes.join(', ')}`;
  updateGrid();
}

function updateGrid(){
  const ts = Date.now();
  qs('#grid').src = `grid?z=${z}&y=${y}&x=${x}&_=${ts}`;
  qs('#indices').innerText = `Z=${z} | Y=${y} | X=${x}`;
}

qs('#zidx').addEventListener('change', () => { z = clamp(qs('#zidx').value, 0, meta.dims[0]-1); updateGrid(); });
qs('#yidx').addEventListener('change', () => { y = clamp(qs('#yidx').value, 0, meta.dims[1]-1); updateGrid(); });
qs('#xidx').addEventListener('change', () => { x = clamp(qs('#xidx').value, 0, meta.dims[2]-1); updateGrid(); });

qs('#shuffle').addEventListener('click', async () => {
  try {
    const r = await fetch(`rand_indices?z=${z}&y=${y}&x=${x}&_=${Date.now()}`, {cache:'no-store'});
    const res = await r.json();
    z = res.z; y = res.y; x = res.x;
    qs('#zidx').value = z; qs('#yidx').value = y; qs('#xidx').value = x;
    updateGrid();
  } catch(e){ console.error(e); }
});

qs('#reset').addEventListener('click', () => {
  z = meta.mid.z; y = meta.mid.y; x = meta.mid.x;
  qs('#zidx').value = z; qs('#yidx').value = y; qs('#xidx').value = x;
  updateGrid();
});

qs('#savefig').addEventListener('click', async () => {
  const r = await fetch(`save_grid?z=${z}&y=${y}&x=${x}`, {cache:'no-store'});
  const res = await r.json();
  qs('#savehint').innerText = res.path ? `Saved: ${res.path}` : (res.error || 'Save failed');
});

loadMeta();
</script>
</body>
</html>
"""


def _safe_bounds(n: int) -> tuple[int, int]:
    if n <= 10:
        return 0, n
    lo, hi = 5, n - 5
    if hi <= lo:
        return 0, n
    return lo, hi


def _rgba_overlay(mask2d: np.ndarray) -> np.ndarray:
    h, w = mask2d.shape
    out = np.zeros((h, w, 4), dtype=np.float32)
    out[mask2d == 1] = [1.0, 0.0, 0.0, 0.55]
    out[mask2d == 2] = [0.0, 1.0, 0.0, 0.55]
    out[mask2d == 3] = [0.0, 0.0, 1.0, 0.55]
    return out


def _figure_pair_grid(X: np.ndarray, Ygt: np.ndarray, Ypr: np.ndarray, z: int, y: int, x: int) -> bytes:
    fig = Figure(figsize=(12, 7), dpi=110)
    axs = fig.subplots(2, 3)
    # Normalize indices
    Dz, Dy, Dx = X.shape
    z = max(0, min(Dz - 1, int(z)))
    y = max(0, min(Dy - 1, int(y)))
    x = max(0, min(Dx - 1, int(x)))

    planes = [
        ('Axial (Z)', X[z, :, :], Ygt[z, :, :], Ypr[z, :, :], z),
        ('Coronal (Y)', X[:, y, :], Ygt[:, y, :], Ypr[:, y, :], y),
        ('Sagittal (X)', X[:, :, x], Ygt[:, :, x], Ypr[:, :, x], x),
    ]

    for c, (title, x2d, gt2d, pr2d, idx) in enumerate(planes):
        vmin, vmax = np.percentile(x2d, [5, 95]).astype(float)
        if not np.isfinite(vmin) or not np.isfinite(vmax) or vmin == vmax:
            vmin, vmax = float(np.min(x2d)), float(np.max(x2d) if np.max(x2d) != np.min(x2d) else np.min(x2d) + 1.0)
        # Row 0: GT
        ax = axs[0, c]; ax.axis('off')
        ax.imshow(x2d, cmap='gray', vmin=vmin, vmax=vmax, interpolation='nearest')
        ax.imshow(_rgba_overlay(gt2d), interpolation='nearest')
        ax.set_title(f"{title} – GT (slice {idx})", fontsize=9)
        # Row 1: Pred
        ax = axs[1, c]; ax.axis('off')
        ax.imshow(x2d, cmap='gray', vmin=vmin, vmax=vmax, interpolation='nearest')
        ax.imshow(_rgba_overlay(pr2d), interpolation='nearest')
        ax.set_title(f"{title} – Pred", fontsize=9)

    fig.tight_layout(pad=0.5)
    buf = io.BytesIO()
    FigureCanvas(fig).print_png(buf)
    return buf.getvalue()


class _Handler(BaseHTTPRequestHandler):
    X = None
    Ygt = None
    Ypr = None
    figs_dir = None
    ALLOW_SAVE = True

    def log_message(self, fmt, *args):
        pass

    def _send(self, code, ctype="text/html; charset=utf-8", body=b"", headers=None):
        self.send_response(code)
        self.send_header("Content-Type", ctype)
        self.send_header("Cache-Control", "no-store")
        if headers:
            for k, v in headers.items():
                self.send_header(k, v)
        self.end_headers()
        if body:
            self.wfile.write(body)

    def do_GET(self):
        parsed = urllib.parse.urlparse(self.path)
        path = parsed.path.lstrip("/")
        q = urllib.parse.parse_qs(parsed.query)

        if path in ("", "index.html"):
            return self._send(200, body=_HTML.encode('utf-8'))

        if path == 'meta':
            Dz, Dy, Dx = self.X.shape
            mid = {
                'z': Dz // 2,
                'y': Dy // 2,
                'x': Dx // 2,
            }
            return self._send(200, 'application/json', json.dumps({
                'dims': [Dz, Dy, Dx],
                'classes': ['Background', 'CSF', 'Gray Matter', 'White Matter'],
                'mid': mid,
            }).encode('utf-8'))

        if path == 'grid':
            Dz, Dy, Dx = self.X.shape
            try:
                z = int(q.get('z', [Dz//2])[0]); y = int(q.get('y', [Dy//2])[0]); x = int(q.get('x', [Dx//2])[0])
            except Exception:
                z, y, x = Dz//2, Dy//2, Dx//2
            try:
                png = _figure_pair_grid(self.X, self.Ygt, self.Ypr, z, y, x)
                return self._send(200, 'image/png', png)
            except Exception as e:
                return self._send(500, 'application/json', json.dumps({'error': str(e)}).encode('utf-8'))

        if path == 'rand_indices':
            Dz, Dy, Dx = self.X.shape
            def rand_idx(n, cur):
                lo, hi = _safe_bounds(n)
                if hi <= lo:
                    return max(0, min(n-1, (cur+1) % max(1, n)))
                import random as _r
                tries = 0
                while True:
                    v = _r.randint(lo, max(lo, hi-1))
                    if v != cur or tries > 8:
                        return v
                    tries += 1
            try:
                cz = int(q.get('z', [Dz//2])[0]); cy = int(q.get('y', [Dy//2])[0]); cx = int(q.get('x', [Dx//2])[0])
            except Exception:
                cz, cy, cx = Dz//2, Dy//2, Dx//2
            payload = {'z': rand_idx(Dz, cz), 'y': rand_idx(Dy, cy), 'x': rand_idx(Dx, cx)}
            return self._send(200, 'application/json', json.dumps(payload).encode('utf-8'))

        if path == 'save_grid':
            if not getattr(self, 'ALLOW_SAVE', True):
                return self._send(403, 'application/json', json.dumps({'error': 'Save disabled'}).encode('utf-8'))
            Dz, Dy, Dx = self.X.shape
            try:
                z = int(q.get('z', [Dz//2])[0]); y = int(q.get('y', [Dy//2])[0]); x = int(q.get('x', [Dx//2])[0])
            except Exception:
                z, y, x = Dz//2, Dy//2, Dx//2
            try:
                png = _figure_pair_grid(self.X, self.Ygt, self.Ypr, z, y, x)
                os.makedirs(self.figs_dir, exist_ok=True)
                path = os.path.join(self.figs_dir, 'webview_infer_grid.png')
                with open(path, 'wb') as f:
                    f.write(png)
                return self._send(200, 'application/json', json.dumps({'path': path}).encode('utf-8'))
            except Exception as e:
                return self._send(500, 'application/json', json.dumps({'error': str(e)}).encode('utf-8'))

        return self._send(404, body=b'Not Found')


def launch_web_inference_viewer(X_sample: np.ndarray,
                                Y_pred: np.ndarray,
                                Y_gt: np.ndarray,
                                figs_dir: str | None = None,
                                host: str = '127.0.0.1',
                                port: int = 0,
                                open_browser: bool = True,
                                allow_save: bool = True) -> str:
    X = np.asarray(X_sample)
    Yp = np.asarray(Y_pred)
    Yg = np.asarray(Y_gt)
    if X.ndim != 3:
        raise ValueError(f"X must be 3D [D,H,W], got {X.shape}")
    if Yp.shape != X.shape or Yg.shape != X.shape:
        raise ValueError("Y_pred and Y_gt must match X shape")
    if figs_dir is None:
        figs_dir = os.path.join(os.path.dirname(__file__), 'saved.ckpt', 'Figs')

    handler = type('_InferHandler', (_Handler,), {})
    handler.X = X
    handler.Ygt = Yg
    handler.Ypr = Yp
    handler.figs_dir = figs_dir
    handler.ALLOW_SAVE = bool(allow_save)

    httpd = ThreadingHTTPServer((host, port), handler)
    url = f"http://{host}:{httpd.server_port}/"
    t = threading.Thread(target=httpd.serve_forever, daemon=True)
    t.start()
    print(f"Inference viewer at {url}")
    if open_browser:
        try:
            webbrowser.open(url)
        except Exception:
            pass
    return url




from __future__ import annotations
from typing import Optional, Sequence, Tuple
import os
import glob
import numpy as np

try:
    import SimpleITK as sitk  # type: ignore
    _HAS_SITK = True
except Exception:
    _HAS_SITK = False

try:
    from skimage.exposure import equalize_adapthist as _clahe, match_histograms as _match_hist  # type: ignore
    _HAS_SKIMAGE = True
except Exception:
    _HAS_SKIMAGE = False

try:
    import nibabel as nib  # type: ignore
    _HAS_NIB = True
except Exception:
    _HAS_NIB = False


def clahe3d(x: np.ndarray, mask: Optional[np.ndarray] = None, *, clip_limit: float = 0.01, kernel_size: Optional[Sequence[int]] = None, nbins: int = 256) -> np.ndarray:
    """3D CLAHE on volume x; if mask given, apply and return x with only masked voxels equalized.

    x: float32 volume [D,H,W]
    """
    if not _HAS_SKIMAGE:
        return x
    if kernel_size is None:
        kernel_size = (8, 32, 32)
    # skimage equalize_adapthist expects values in [0,1]
    x_min, x_max = float(np.min(x)), float(np.max(x))
    xr = (x - x_min) / (x_max - x_min + 1e-6)
    xe = _clahe(xr, kernel_size=tuple(kernel_size), clip_limit=float(clip_limit), nbins=int(nbins))
    # Rescale back to original scale
    xe = xe * (x_max - x_min) + x_min
    if mask is None:
        return xe.astype(x.dtype, copy=False)
    out = x.copy()
    out[mask > 0.5] = xe[mask > 0.5]
    return out.astype(x.dtype, copy=False)


def hist_match3d(x: np.ndarray, ref: np.ndarray, mask: Optional[np.ndarray] = None, ref_mask: Optional[np.ndarray] = None) -> np.ndarray:
    """3D histogram matching of x to ref.

    - If no masks, prefer SimpleITK's HistogramMatching (fast, robust).
    - If masks provided, use scikit-image match_histograms on the masked vectors and assign back.
    """
    if mask is None and ref_mask is None and _HAS_SITK:
        try:
            mov = sitk.GetImageFromArray(x.astype(np.float32))
            ref_img = sitk.GetImageFromArray(ref.astype(np.float32))
            hm = sitk.HistogramMatchingImageFilter()
            hm.SetNumberOfHistogramLevels(2048)
            hm.SetNumberOfMatchPoints(100)
            hm.SetThresholdAtMeanIntensity(True)
            out = hm.Execute(mov, ref_img)
            return sitk.GetArrayFromImage(out).astype(x.dtype, copy=False)
        except Exception:
            pass
    if not _HAS_SKIMAGE:
        # Fallback: if scikit-image not available, try SimpleITK global HM even if masks were requested
        if _HAS_SITK:
            try:
                mov = sitk.GetImageFromArray(x.astype(np.float32))
                ref_img = sitk.GetImageFromArray(ref.astype(np.float32))
                hm = sitk.HistogramMatchingImageFilter()
                hm.SetNumberOfHistogramLevels(2048)
                hm.SetNumberOfMatchPoints(100)
                hm.SetThresholdAtMeanIntensity(True)
                out = hm.Execute(mov, ref_img)
                return sitk.GetArrayFromImage(out).astype(x.dtype, copy=False)
            except Exception:
                pass
        return x
    if mask is None and ref_mask is None:
        return _match_hist(x, ref).astype(x.dtype, copy=False)
    # Mask-aware path: match masked moving voxels to masked reference voxels (different shapes ok)
    m = (mask > 0.5) if mask is not None else np.ones_like(x, dtype=bool)
    if ref_mask is not None:
        rsel = ref[ref_mask > 0.5]
    else:
        rsel = ref.ravel()
    xsel = x[m]
    xmatched = _match_hist(xsel, rsel).astype(x.dtype, copy=False)
    out = x.copy()
    out[m] = xmatched
    return out


def n4_bias_correct(x: np.ndarray, mask: Optional[np.ndarray] = None, shrink_factor: int = 2, num_iterations: int = 50) -> np.ndarray:
    """Apply N4 bias field correction (SimpleITK) on 3D volume; returns float32 array.
    If mask provided, constrains correction to head/brain region.
    """
    if not _HAS_SITK:
        return x
    try:
        img = sitk.GetImageFromArray(x.astype(np.float32))
        if mask is not None:
            m = sitk.GetImageFromArray((mask.astype(np.uint8)))
        else:
            m = sitk.OtsuThreshold(img, 0, 1)
        corrector = sitk.N4BiasFieldCorrectionImageFilter()
        corrector.SetMaximumNumberOfIterations([num_iterations])
        corrector.SetShrinkFactor(shrink_factor)
        out = corrector.Execute(img, m)
        return sitk.GetArrayFromImage(out).astype(np.float32, copy=False)
    except Exception:
        return x


def ahe3d_sitk(x: np.ndarray, alpha: float = 1.0, beta: float = 1.0) -> np.ndarray:
    """Adaptive Histogram Equalization using SimpleITK to mirror 2D pipeline; returns float32 numpy array."""
    if not _HAS_SITK:
        return x
    img = sitk.GetImageFromArray(x.astype(np.float32))
    ahe = sitk.AdaptiveHistogramEqualizationImageFilter()
    ahe.SetAlpha(float(alpha))
    ahe.SetBeta(float(beta))
    out = ahe.Execute(img)
    return sitk.GetArrayFromImage(out).astype(np.float32, copy=False)


def load_ibsr_reference_from_env_or_meta(meta) -> Optional[np.ndarray]:
    """Load an IBSR reference volume (RAS), apply AHE to serve as HM anchor; returns 3D numpy or None."""
    # Meta-provided ref path takes precedence
    try:
        from .config import get as cfg  # local import to avoid cycles
    except Exception:
        cfg = lambda m, k, d=None: d  # type: ignore

    ref_path = None
    try:
        ref_path = cfg(meta, "preproc.ibsr_ref")
    except Exception:
        ref_path = None

    # Env IBSR_ROOT fallback
    if not ref_path:
        root = os.environ.get("IBSR_ROOT")
        if root and os.path.isdir(root):
            # Prefer a canonical subject; else any *_ana.nii.gz
            cand = os.path.join(root, "IBSR_01", "IBSR_01_ana.nii.gz")
            if os.path.exists(cand):
                ref_path = cand
            else:
                hits = glob.glob(os.path.join(root, "IBSR_*", "IBSR_*_ana.nii.gz"))
                ref_path = hits[0] if hits else None
    # Hard fallback: None
    if not ref_path or not _HAS_NIB:
        return None
    try:
        img = nib.load(ref_path)
        img = nib.as_closest_canonical(img)
        arr = img.get_fdata().astype(np.float32)
        if arr.ndim == 4:
            arr = arr[..., 0]
        # Apply AHE to mirror prior pipeline
        arr = ahe3d_sitk(arr, alpha=1.0, beta=1.0)
        return arr
    except Exception:
        return None


def load_ibsr_reference_array_and_zooms(meta) -> Optional[tuple[np.ndarray, tuple[float, float, float]]]:
    """Load IBSR ref array (AHE-processed) and its zooms (spacing)."""
    if not _HAS_NIB:
        return None
    try:
        # Reuse path resolution from helper above
        from .config import get as cfg
        ref_path = cfg(meta, "preproc.ibsr_ref")
        if not ref_path:
            root = os.environ.get("IBSR_ROOT")
            if root and os.path.isdir(root):
                cand = os.path.join(root, "IBSR_01", "IBSR_01_ana.nii.gz")
                if os.path.exists(cand):
                    ref_path = cand
                else:
                    hits = glob.glob(os.path.join(root, "IBSR_*", "IBSR_*_ana.nii.gz"))
                    ref_path = hits[0] if hits else None
        if not ref_path:
            return None
        img = nib.load(ref_path)
        img = nib.as_closest_canonical(img)
        arr = img.get_fdata().astype(np.float32)
        if arr.ndim == 4:
            arr = arr[..., 0]
        arr = ahe3d_sitk(arr, alpha=1.0, beta=1.0)
        zooms = tuple(float(z) for z in img.header.get_zooms()[:3])
        return arr, zooms
    except Exception:
        return None


def load_ibsr_reference_array_mask_and_zooms(meta) -> Optional[tuple[np.ndarray, Optional[np.ndarray], tuple[float, float, float]]]:
    """Load IBSR ref array (AHE-processed), an optional reference brain mask, and zooms.

    Meta keys:
      - preproc.ibsr_ref: path to IBSR reference anatomical NIfTI
      - preproc.ibsr_ref_mask (optional): path to a brain mask in the same space
    """
    if not _HAS_NIB:
        return None
    try:
        from .config import get as cfg
        ref_path = cfg(meta, "preproc.ibsr_ref")
        mask_path = cfg(meta, "preproc.ibsr_ref_mask")
        if not ref_path:
            root = os.environ.get("IBSR_ROOT")
            if root and os.path.isdir(root):
                cand = os.path.join(root, "IBSR_01", "IBSR_01_ana.nii.gz")
                if os.path.exists(cand):
                    ref_path = cand
                else:
                    hits = glob.glob(os.path.join(root, "IBSR_*", "IBSR_*_ana.nii.gz"))
                    ref_path = hits[0] if hits else None
        if not ref_path:
            return None
        img = nib.load(ref_path)
        img = nib.as_closest_canonical(img)
        arr = img.get_fdata().astype(np.float32)
        if arr.ndim == 4:
            arr = arr[..., 0]
        arr = ahe3d_sitk(arr, alpha=1.0, beta=1.0)
        zooms = tuple(float(z) for z in img.header.get_zooms()[:3])
        ref_mask = None
        if mask_path and os.path.exists(mask_path):
            mimg = nib.load(mask_path)
            mimg = nib.as_closest_canonical(mimg)
            m = mimg.get_fdata()
            if m.ndim == 4:
                m = m[..., 0]
            # Consider any nonzero as brain
            ref_mask = (m > 0).astype(np.uint8)
        return arr, ref_mask, zooms
    except Exception:
        return None


def load_nfbs_reference_array_and_zooms(meta) -> Optional[tuple[np.ndarray, tuple[float, float, float]]]:
    """Load NFBS reference array (optionally AHE-processed) and zooms.

    Meta key: preproc.nfbs_ref
    """
    if not _HAS_NIB:
        return None
    try:
        from .config import get as cfg
        ref_path = cfg(meta, "preproc.nfbs_ref")
        if not ref_path or not os.path.exists(ref_path):
            return None
        img = nib.load(ref_path)
        img = nib.as_closest_canonical(img)
        arr = img.get_fdata().astype(np.float32)
        if arr.ndim == 4:
            arr = arr[..., 0]
        # Keep same AHE anchor behavior for consistency
        arr = ahe3d_sitk(arr, alpha=1.0, beta=1.0)
        zooms = tuple(float(z) for z in img.header.get_zooms()[:3])
        return arr, zooms
    except Exception:
        return None


def robust_winsorize(x: np.ndarray, mask: Optional[np.ndarray] = None, lo: float = 1.0, hi: float = 99.0) -> Tuple[np.ndarray, float, float]:
    """Clamp x to [lo,hi] percentiles computed over mask (or all voxels if mask None).

    Returns (winsorized_x, p_lo, p_hi).
    """
    try:
        if mask is not None:
            vals = x[mask > 0]
            if vals.size < 10:
                vals = x.ravel()
        else:
            vals = x.ravel()
        plo, phi = np.percentile(vals, [float(lo), float(hi)])
        if phi > plo:
            xo = np.clip(x, plo, phi)
            return xo, float(plo), float(phi)
        return x, float(plo), float(phi)
    except Exception:
        return x, float(np.min(x)), float(np.max(x))


def percentile_linear_map_to_ref(x: np.ndarray,
                                 mask: Optional[np.ndarray],
                                 ref: np.ndarray,
                                 ref_mask: Optional[np.ndarray],
                                 lo: float = 1.0,
                                 hi: float = 99.0) -> Tuple[np.ndarray, Tuple[float, float, float, float]]:
    """Linearly map x's [lo,hi] masked percentiles to ref's [lo,hi] masked percentiles.

    Returns (mapped_x, (x_lo, x_hi, r_lo, r_hi)).
    """
    try:
        xv = x[mask > 0] if mask is not None else x.ravel()
        rv = ref[ref_mask > 0] if ref_mask is not None else ref.ravel()
        xlo, xhi = np.percentile(xv, [float(lo), float(hi)])
        rlo, rhi = np.percentile(rv, [float(lo), float(hi)])
        if (xhi > xlo) and (rhi > rlo):
            y = (x - xlo) / (xhi - xlo)
            y = y * (rhi - rlo) + rlo
            return y.astype(np.float32, copy=False), (float(xlo), float(xhi), float(rlo), float(rhi))
        return x, (float(xlo), float(xhi), float(rlo), float(rhi))
    except Exception:
        return x, (float(np.min(x)), float(np.max(x)), float(np.min(ref)), float(np.max(ref)))


## Removed experimental intensity mapping helpers (not used in current pipeline)



from __future__ import annotations
import os
import sys
from typing import Optional, Any, Tuple


def _ensure_mednet_path() -> None:
    """Insert a vendored MEDNet path relative to this repo (standalone).

    Tries, in order:
      - <repo>/LongitudinalAnalysisAIIMS.pypi/MEDNet.pypi
      - <repo>/LongitudinalAnalysisAIIMS.pypi/MEDNet
      - <repo>/LongitudinalAnalysisAIIMS.pypi/vendor/MEDNet.pypi
    """
    here = os.path.abspath(os.path.dirname(__file__))
    candidates = [
        os.path.abspath(os.path.join(here, "..", "MEDNet.pypi")),   # vendored as MEDNet.pypi/
        os.path.abspath(os.path.join(here, "..", "MEDNet")),        # vendored as MEDNet/
        os.path.abspath(os.path.join(here, "..", "vendor", "MEDNet.pypi")),
    ]
    for p in candidates:
        if os.path.isdir(p) and p not in sys.path:
            sys.path.insert(0, p)
            break


def build_model(model_type: str,
                input_size: int,
                n_classes: int,
                loss: str,
                class_weights: Optional[Any],
                wave: Optional[str] = None,
                compile_flag: bool = True):
    _ensure_mednet_path()
    from MEDNet.model_factory import ModelFactory
    # Create model (wrapped G3D provides loss/metrics)
    model = ModelFactory.create_model(
        model_type=model_type,
        input_size=input_size,
        n_classes=n_classes,
        loss=loss,
        class_weights=class_weights,
        build=True,
        compile=compile_flag,
        wave=wave,
    )
    return model


def load_weights_best_effort(model, checkpoint_path: str) -> None:
    """Strict weight load: raise if missing or mismatched."""
    if not checkpoint_path or not os.path.exists(checkpoint_path):
        raise FileNotFoundError(f"Checkpoint not found: {checkpoint_path}")
    model.load_weights(checkpoint_path)


def build_models_from_meta(meta) -> Tuple[object, object]:
    """Build skull-strip and segmentation models from meta configuration.

    Returns (ss_model, seg_model). Does not alter any preprocessing/dataflow.
    """
    # Import here to avoid circulars
    from .config import get as cfg

    # Inference sizes
    psize_default = int(cfg(meta, "infer.patch_size", 256))
    psize_ss = int(cfg(meta, "infer.patch_size_ss", psize_default))
    psize_seg = int(cfg(meta, "infer.patch_size_seg", psize_default))

    # SS model
    ss_cfg = cfg(meta, "models.ss") or {}
    if not ss_cfg or not ss_cfg.get("checkpoint"):
        raise ValueError("models.ss.checkpoint is required in meta")
    n_classes_ss = int(ss_cfg.get("n_classes", 2))
    loss_ss = ss_cfg.get("loss", "focal")
    ckpt_ss = ss_cfg["checkpoint"]
    type_ss = ss_cfg.get("type", "uwavevitnet")
    wave_ss = ss_cfg.get("wave", None)
    class_weights_ss = [0.1] * n_classes_ss if "focal" in (loss_ss or "").lower() else None
    ss_model = build_model(type_ss, input_size=psize_ss, n_classes=n_classes_ss, loss=loss_ss, class_weights=class_weights_ss, wave=wave_ss, compile_flag=True)
    load_weights_best_effort(ss_model, ckpt_ss)

    # Seg model
    seg_cfg = cfg(meta, "models.seg") or {}
    if not seg_cfg or not seg_cfg.get("checkpoint"):
        raise ValueError("models.seg.checkpoint is required in meta")
    n_classes_seg = int(seg_cfg.get("n_classes", 4))
    loss_seg = seg_cfg.get("loss", "focal")
    ckpt_seg = seg_cfg["checkpoint"]
    type_seg = seg_cfg.get("type", "uwavevitnet")
    wave_seg = seg_cfg.get("wave", None)
    class_weights_seg = [0.1] * n_classes_seg if "focal" in (loss_seg or "").lower() else None
    seg_model = build_model(type_seg, input_size=psize_seg, n_classes=n_classes_seg, loss=loss_seg, class_weights=class_weights_seg, wave=wave_seg, compile_flag=True)
    load_weights_best_effort(seg_model, ckpt_seg)

    return ss_model, seg_model



from __future__ import annotations
import os
from typing import Tuple, Optional, Dict
import numpy as np
import nibabel as nib


def load_nifti_as_ras(path: str) -> Tuple[np.ndarray, np.ndarray, nib.nifti1.Nifti1Header, Tuple[float, float, float]]:
    """Load NIfTI, reorient to closest canonical (RAS), return data [D,H,W,1] float32.

    Returns: (data4d, affine, header, zooms)
    """
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    img = nib.load(path)
    img_c = nib.as_closest_canonical(img)
    data = img_c.get_fdata(dtype=np.float32)
    # Ensure 3D
    if data.ndim == 4:
        # take first channel if necessary
        data = data[..., 0]
    if data.ndim != 3:
        raise ValueError(f"Expected 3D volume, got shape {data.shape}")
    data4d = data[..., None]
    hdr = img_c.header.copy()
    zooms = tuple(float(z) for z in hdr.get_zooms()[:3])
    return data4d, img_c.affine, hdr, zooms


def save_nifti(data3d: np.ndarray, affine: np.ndarray, header: nib.nifti1.Nifti1Header, out_path: str) -> None:
    os.makedirs(os.path.dirname(out_path), exist_ok=True)
    # Preserve header dtype carefully
    hdr = header.copy()
    # Adjust dtype based on data type
    img = nib.Nifti1Image(data3d, affine, header=hdr)
    nib.save(img, out_path)


def zscore_normalize(x: np.ndarray, eps: float = 1e-6) -> np.ndarray:
    mu = np.mean(x)
    sigma = np.std(x)
    if sigma < eps:
        return x * 0.0
    return (x - mu) / (sigma + eps)


def percentile_clip(x: np.ndarray, lo: float = 0.5, hi: float = 99.5) -> np.ndarray:
    lo_v = np.percentile(x, lo)
    hi_v = np.percentile(x, hi)
    x = np.clip(x, lo_v, hi_v)
    # Rescale to 0-1
    if hi_v > lo_v:
        x = (x - lo_v) / (hi_v - lo_v)
    return x


def maybe_centerpad_h_axis_to_cube(vol: np.ndarray) -> np.ndarray:
    """If the volume is 256x128x256 (IBSR), zero-pad H by 64 on both sides to 256.
    vol shape: [D,H,W,1]
    """
    if vol.ndim != 4 or vol.shape[-1] != 1:
        raise ValueError("Expected [D,H,W,1]")
    D, H, W, C = vol.shape
    if (D, H, W) == (256, 128, 256):
        pad_cfg = ((0, 0), (64, 64), (0, 0), (0, 0))
        return np.pad(vol, pad_cfg, mode="constant")
    return vol


def load_raw_then_crop_then_ras(path: str,
                                crop: Optional[Dict[str, Optional[Tuple[int, int]]]] = None) -> Tuple[np.ndarray, np.ndarray, nib.nifti1.Nifti1Header, Tuple[float, float, float]]:
    """Load NIfTI in raw orientation, apply manual cropping slices, then canonicalize to RAS.

    crop: dict with optional keys 'z','y','x' each as (start, end) or None.
    Returns data4d [D,H,W,1] float32, affine, header, zooms.
    """
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    img = nib.load(path)
    nii = img
    if crop:
        # Note: NIfTI array order is (X, Y, Z). Apply slices as (sx, sy, sz).
        zc = crop.get('z') if 'z' in crop else None
        yc = crop.get('y') if 'y' in crop else None
        xc = crop.get('x') if 'x' in crop else None
        sz = slice(None) if (zc is None or zc == []) else slice(int(zc[0]), int(zc[1]))
        sy = slice(None) if (yc is None or yc == []) else slice(int(yc[0]), int(yc[1]))
        sx = slice(None) if (xc is None or xc == []) else slice(int(xc[0]), int(xc[1]))
        try:
            nii = img.slicer[sx, sy, sz]
        except Exception:
            # Fallback: crop array then wrap; affine will be approximate (no offset)
            arr = img.get_fdata().astype(np.float32)
            arr = arr[sx, sy, sz]
            nii = nib.Nifti1Image(arr, img.affine, header=img.header)
    # Canonicalize after manual crop
    nii_ras = nib.as_closest_canonical(nii)
    data = nii_ras.get_fdata(dtype=np.float32)
    if data.ndim == 4:
        data = data[..., 0]
    data4d = data[..., None]
    hdr = nii_ras.header.copy()
    zooms = tuple(float(z) for z in hdr.get_zooms()[:3])
    return data4d, nii_ras.affine, hdr, zooms

def load_label_as_ras(path: str) -> Tuple[np.ndarray, np.ndarray, nib.nifti1.Nifti1Header, Tuple[float, float, float]]:
    """Load label NIfTI as closest canonical (RAS); return int16 3D array.

    Returns: (labels3d, affine, header, zooms)
    """
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    img = nib.load(path)
    img_c = nib.as_closest_canonical(img)
    data = img_c.get_fdata()
    if data.ndim == 4:
        data = data[..., 0]
    labels = np.rint(data).astype(np.int16)
    hdr = img_c.header.copy()
    zooms = tuple(float(z) for z in hdr.get_zooms()[:3])
    return labels, img_c.affine, hdr, zooms



import tensorflow as tf

@tf.keras.utils.register_keras_serializable()
class Dice(tf.keras.metrics.Metric):
    """ Computes the Dice score for a single class ID using tf.keras.losses.Dice(reduction='none').
        Instantiate separately for each class to get per-class Dice.
    
        Should work similar to IoU metric of tf by using a target_class_id   

        tf-nd-utils: Multidimensional utility layers in TensorFlow.
        Copyright (C) 2025 Kishore Kumar Tarafdar

        This program is free software: you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation, either version 3 of the License, or
        (at your option) any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
    ---kkt@19Jun2025"""
    def __init__(self, target_class_ids, name=None, **kwargs):
        if not isinstance(target_class_ids, (list, tuple)) or len(target_class_ids) != 1:
            raise ValueError("For per-class Dice, target_class_ids must be a list with exactly one class index.")

        self.class_id = target_class_ids[0]
        name = name or f"dice_class_{self.class_id}"

        super().__init__(name=name, **kwargs)
        self.loss_fn = tf.keras.losses.Dice(reduction='none')
        self.dice_sum = self.add_weight(name="dice_sum", initializer="zeros")
        self.count = self.add_weight(name="count", initializer="zeros")

    def update_state(self, y_true, y_pred, sample_weight=None):
        ## new patch for sparse case
        # y_pred: [batch, ..., n_classes]
        # y_true: [batch, ...] (sparse) or [batch, ..., n_classes] (one-hot)
        n_classes = tf.shape(y_pred)[-1]
        # Handle [B, D, H, W, 1] case
        if y_true.shape.rank == y_pred.shape.rank and y_true.shape[-1] == 1:
            y_true = tf.squeeze(y_true, axis=-1)
        # One-hot encode if needed
        if y_true.shape.rank == y_pred.shape.rank - 1:
            y_true = tf.one_hot(tf.cast(y_true, tf.int32), depth=n_classes)
        else:
            tf.debugging.assert_equal(
                tf.shape(y_true)[-1], n_classes,
                message="y_true and y_pred should have same number of channels"
            )
        #---new patch sparse case

        y_true_cls = tf.expand_dims(y_true[..., self.class_id], axis=-1)
        y_pred_cls = tf.expand_dims(y_pred[..., self.class_id], axis=-1)

        dice_loss = self.loss_fn(y_true_cls, y_pred_cls)  # shape: (batch,) or scalar
        dice_score = 1.0 - dice_loss

        #---new patch
        if sample_weight is not None:
            dice_score = dice_score * tf.cast(sample_weight, self.dtype)

        batch_score = tf.reduce_sum(dice_score)
        batch_count = tf.cast(tf.size(dice_score), tf.float32)

        self.dice_sum.assign_add(batch_score)
        self.count.assign_add(batch_count)

    def result(self):
        return self.dice_sum / (self.count + 1e-7)

    def reset_states(self):
        self.dice_sum.assign(0.0)
        self.count.assign(0.0)

    def get_config(self):
        config = super().get_config()
        config.update({
            "target_class_ids": [self.class_id]
        })
        return config



"""
Model Factory for 3D Brain Segmentation
---------------------------------------
Centralizes model creation for MEDNet 3D architectures.

It prefers the MEDNet package (renamed from MEDNet_nightly).
"""

import os
 
import tensorflow as tf
Gφψ3D_ViT = None  # default if not available; prefer nightly import below

# Prefer renamed stable package (formerly MEDNet_nightly)
from MEDNet.G3D import Gφψ3D, G3D, configs
from MEDNet.UNet3D import UNet3D
try:
    from MEDNet.UNet3Ddwtattn import UNet3Ddwtattn  # optional
except Exception:
    UNet3Ddwtattn = None  # type: ignore
try:
    from MEDNet.UNet3Dattention import UNet3D_with_attention as UNet3Dattention  # optional
except Exception:
    UNet3Dattention = None  # type: ignore
try:
    from MEDNet.G3Dattention import Gφψ3D_attention as G3Dattention  # optional
except Exception:
    G3Dattention = None  # type: ignore
try:
    from MEDNet.UNet3DViT import UNet3DViT  # optional
except Exception:  # pragma: no cover - optional
    UNet3DViT = None  # type: ignore
try:
    from MEDNet.G3DViT import Gφψ3D_ViT  # optional
except Exception:  # pragma: no cover - optional
    Gφψ3D_ViT = None  # type: ignore
try:
    from MEDNet.G3DwaveViT import Gφψ3D_waveViT  # optional
except Exception:  # pragma: no cover - optional
    Gφψ3D_waveViT = None  # type: ignore
try:
    # Module renamed: FDSTUNet3D.py -> ShearMEDNet3D.py; factory keeps the same symbol name
    from MEDNet.ShearMEDNet3D import FDSTUNet3D  # optional
except Exception:  # pragma: no cover - optional
    FDSTUNet3D = None  # type: ignore
try:
    # New: ShearMED UNet with FDST-native ViT bottleneck (deterministic)
    from MEDNet.ShearMEDFDSTViTNet3D import ShearMEDFDSTViTNet3D  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] ShearMEDFDSTViTNet3D import failed:", type(_e).__name__, _e)
    ShearMEDFDSTViTNet3D = None  # type: ignore
try:
    # Module renamed: FDSTCVAEUNet3D.py -> CVShearMEDNet3D.py; factory keeps the same symbol name
    from MEDNet.CVShearMEDNet3D import FDSTCVAEUNet3D  # optional
except Exception:  # pragma: no cover - optional
    FDSTCVAEUNet3D = None  # type: ignore
try:
    # Renamed: FDSTCVAEUNetWaveViT3D.py -> CVShearMEDWaveViTNet3D.py; export alias provided
    from MEDNet.CVShearMEDWaveViTNet3D import CVShearMEDWaveViTNet3D  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] CVShearMEDWaveViTNet3D import failed:", type(_e).__name__, _e)
    CVShearMEDWaveViTNet3D = None  # type: ignore
try:
    from MEDNet.G3DLifting import Gφψ3D as Gφψ3D_Lifting  # optional
except Exception:  # pragma: no cover - optional
    Gφψ3D_Lifting = None  # type: ignore
try:
    from MEDNet.G3D_minimal import Gφψ3D as Gφψ3D_Minimal  # optional
except Exception:  # pragma: no cover - optional
    Gφψ3D_Minimal = None  # type: ignore
try:
    # Module was renamed from CVAEUNet3D.py to CVUNet3D.py; class name unchanged
    from MEDNet.CVUNet3D import CVAEUNet3D  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] CVAEUNet3D import failed:", type(_e).__name__, _e)
    CVAEUNet3D = None  # type: ignore
try:
    # Use the new, single source of truth
    from MEDNet.CVUWaveViTNet3D import CVUWaveViTNet3D  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] CVUWaveViTNet3D import failed:", type(_e).__name__, _e)
    CVUWaveViTNet3D = None  # type: ignore
try:
    # New ShearMED UNet variant with FDSTViT VAE bottleneck
    from MEDNet.CVShearMEDFDSTViTNet3D import FDSTCVAEUNetFDSTViT3D  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] CVShearMEDFDSTViTNet3D import failed:", type(_e).__name__, _e)
    FDSTCVAEUNetFDSTViT3D = None  # type: ignore
try:
    # New UNet with FDSTViT VAE bottleneck (coefficient-domain)
    # Module renamed: CVAEUNetFDSTViT3D.py -> CVUFDSTViTNet3D.py (use the canonical exported alias)
    from MEDNet.CVUFDSTViTNet3D import CVUFDSTViTNet3D  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] CVUFDSTViTNet3D import failed:", type(_e).__name__, _e)
    CVUFDSTViTNet3D = None  # type: ignore
try:
    # Deterministic UNet with WaveletViT 3D bottleneck (feature-return)
    from MEDNet.UWaveViTNet3D import UNetWaveViT3D  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] UNetWaveViTNet3D import failed:", type(_e).__name__, _e)
    UNetWaveViT3D = None  # type: ignore
try:
    # Hybrid: IIR DF-II conv blocks + WaveletViT 3D bottleneck (deterministic)
    from MEDNet.UIIRDFIIWaveViTNet3D import UIIRDFIIWaveViTNet3D  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] UIIRDFIIWaveViTNet3D import failed:", type(_e).__name__, _e)
    UIIRDFIIWaveViTNet3D = None  # type: ignore
try:
    # Use the new DF-II separable IIR UNet variant (module renamed)
    from MEDNet.UIIRDFIINet3D import UIIRDFIINet3D as UIIRNet3D  # optional alias
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] UIIRDFIINet3D import failed:", type(_e).__name__, _e)
    UIIRNet3D = None  # type: ignore
try:
    from MEDNet.G3DIIRWaveViT import Gφψ3D_iirWaveViT  # optional
except Exception as _e:  # pragma: no cover - optional
    print("[model_factory] G3DIIRWaveViT import failed:", type(_e).__name__, _e)
    Gφψ3D_iirWaveViT = None  # type: ignore
from MEDNet.conv_systems.vanilla_conv import default_conv_block as _default_conv_block
from MEDNet.bottlenecks.vanilla_bottleneck import default_bottleneck as _default_bottleneck

# =============================================================================
# Model Configurations
# =============================================================================

# Config dictionary for different model configurations
configs = configs or {  # type: ignore[name-defined]
    '4567': (16, 32, 64, 128),
}

COMPONENT_REGISTRY = {
    'vanilla_conv': _default_conv_block,
    'vanilla_bottleneck': _default_bottleneck,
}

def get_conv_fn(conv_type):
    """Get convolution function by type"""
    if conv_type not in COMPONENT_REGISTRY:
        raise ValueError(f"Unknown conv_type: {conv_type}. Available: {list(COMPONENT_REGISTRY.keys())}")
    return COMPONENT_REGISTRY[conv_type]

def get_bottleneck_fn(bottleneck_type):
    """Get bottleneck function by type"""
    if bottleneck_type not in COMPONENT_REGISTRY:
        raise ValueError(f"Unknown bottleneck_type: {bottleneck_type}. Available: {list(COMPONENT_REGISTRY.keys())}")
    return COMPONENT_REGISTRY[bottleneck_type]

# =============================================================================
# Model Configurations
# =============================================================================

MODEL_CONFIGS = {
    'G': {
        'name': 'G3D',
        'description': 'G3D model with focal loss for 3D brain segmentation',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    'U': {
        'name': 'UNet3D',
        'description': 'UNet3D model with focal loss for 3D brain segmentation',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    'ufdstvitnet': {
        'name': 'UFDSTViTNet3D',
        'description': 'UFDSTViTNet3D with deterministic FDST-native ViT bottleneck (no VAE)',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    # ViT-backed U-Net
    'uvit': {
        'name': 'UNet3DViT',
        'description': 'UNet3D with Vision Transformer bottleneck',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
    },
    # ViT-backed G3D
    'gvit': {
        'name': 'G3DViT',
        'description': 'G3D with Vision Transformer bottleneck',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
    },
    # Attention U-Net
    'uattn': {
        'name': 'UNet3Dattention',
        'description': 'UNet3D with attention gates on skip connections',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    # Attention G3D
    'gattn': {
        'name': 'G3Dattention',
        'description': 'G3D with attention gates on skip connections',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    # DWT Attention U-Net
    'udwtattn': {
        'name': 'UNet3Ddwtattn',
        'description': 'UNet3D with DWT self-attention',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    # Wavelet-native ViT G3D
    'gwavevit': {
        'name': 'G3DwaveViT',
        'description': 'G3D with wavelet-native ViT bottleneck',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
    },
    # FDST-based UNet3D (renamed from 'fdstunet' -> 'shearnet' -> 'shearmednet')
    'shearmednet': {
        'name': 'FDSTUNet3D',
        'description': 'UNet3D with Fast Discrete Shearlet Transform features',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    # ShearMED UNet with deterministic FDST-native ViT bottleneck
    'shearmedfdstvitnet': {
        'name': 'ShearMEDFDSTViTNet3D',
        'description': 'ShearMED UNet with FDST-native ViT bottleneck (no VAE, coef-domain)',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    # FDST-based UNet3D with CVAE bottleneck
    'cvshearmednet': {
        'name': 'CVShearMEDNet3D',
        'description': 'FDST-UNet with a CVAE bottleneck at the center',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
    # FDST-based UNet3D with WaveViT VAE bottleneck
    'cvshearmedwavevitnet': {
        'name': 'CVShearMEDWaveViTNet3D',
        'description': 'FDST-UNet with WaveViT VAE bottleneck (μ/logσ² over tokens)',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    },
        
    # G3D Lifting with DWT self-attention
    'glifting': {
        'name': 'G3DLifting',
        'description': 'G3D with lifting scheme and DWT self-attention at full resolution',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    }
    ,
    # Minimal G3D variant (direct DWT-based UNet-like)
    'gminimal': {
        'name': 'G3D_minimal',
        'description': 'Minimal G3D (DWT3D/IDWT3D) baseline',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    }
    ,
    # CVAE U-Net (self-contained wrapper)
    'cvunet': {
        'name': 'CVAEUNet3D',
        'description': 'Self-contained CVAE U-Net (KL + focal) built from Encoder/Decoder',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    }
    ,
    # CVAE UNet with WaveViT bottleneck
    'cvuwavevitnet': {
        'name': 'CVUWaveViTNet3D',
        'description': 'UNet3D with WaveViT CV bottleneck (mu/logvar over tokens)',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
    }
    ,
    # ShearMED UNet with FDSTViT VAE bottleneck (coefficient-domain)
    'cvshearmedfdstvitnet': {
        'name': 'CVShearMEDFDSTViTNet3D',
        'description': 'ShearMED UNet with FDSTViT VAE bottleneck (coef-domain, token/feature modes)',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
    }
    ,
    # CVAE UNet with FDSTViT (coefficient-domain) bottleneck
    'cvufdstvitnet': {
        'name': 'CVUFDSTViTNet3D',
        'description': 'UNet3D with FDSTViT VAE bottleneck (coef-domain, z fused into features)',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
    }
    ,
    # UNet3D with WaveletViT 3D bottleneck (deterministic)
    'uwavevitnet': {
        'name': 'UNetWaveViT3D',
        'description': 'UNet3D with 3D Wavelet-ViT bottleneck (feature-domain, no VAE)',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
    }
    ,
    
    # Preferred key for the same hybrid as above (UIIR + WaveViT bottleneck)
    'uiirwavevitnet': {
        'name': 'UIIRDFIIWaveViTNet3D',
        'description': 'UNet3D (IIR DF-II convs) with 3D Wavelet-ViT bottleneck (feature-domain, no VAE)',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
    }
    ,
    # UNet3D variant using IIR3D layers
    'uiir': {
        'name': 'UIIRDFIINet3D',
        'description': 'UNet3D built with IIR3D layers in place of Conv3D',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv',
        'bottleneck_fn': 'vanilla_bottleneck'
    }
    ,
    # Gφψ3D WaveViT variant with IIR3D conv replacements
    'giirwavevit': {
        'name': 'G3DIIRWaveViT',
        'description': 'G3D wavelet-native ViT with IIR3D conv replacements',
        'n_classes': 4,
        'loss': 'focal',
        'config_key': '4567',
        'conv_fn': 'vanilla_conv'
    }
}

# =============================================================================
# Model Factory Class
# =============================================================================

class ModelFactory:
    """
    Factory class for creating 3D segmentation models.

    This factory supports creating different model types with consistent
    interfaces and configurations. New model types can be easily added
    by extending the MODEL_CONFIGS dictionary and adding corresponding
    creation methods.
    """

    @staticmethod
    def create_model(
        model_type,
        input_size=64,
        *,
        n_classes=None,
        loss=None,
        class_weights=None,
        build=True,
        compile=True,
        optimizer=None,
        jit_compile=False,
        show_summary=False,
        wave=None,
    ):
        """
        Create and optionally build/compile the specified model.

        Args:
            model_type: Model type identifier ('G' for G3D, 'U' for UNet3D)
            input_size: Input dimension size (default 64 for patches, 256 for full-size)
            n_classes: Optional override for number of classes
            loss: Optional override for loss name passed to G3D wrapper
            class_weights: Optional class weights passed to G3D wrapper
            build: If True (default), call model.build with appropriate input shape
            compile: If True (default), compile with provided or default optimizer
            optimizer: Optional tf.keras optimizer; defaults to Adam(1e-4)
            jit_compile: Whether to enable XLA JIT at compile time (default False)
            show_summary: If True, print model.summary()
            wave: Wavelet name for G-type models (e.g., 'haar'); ignored for U-type models

        Returns:
            Keras model ready for training/inference

        Raises:
            ValueError: If model_type is not supported
        """
        if model_type not in MODEL_CONFIGS:
            supported_models = list(MODEL_CONFIGS.keys())
            raise ValueError(f"Unknown model type: {model_type}. Supported: {supported_models}")

        config = MODEL_CONFIGS[model_type]
        # Resolve overrides
        resolved_n_classes = n_classes or config['n_classes']
        resolved_loss = loss or config['loss']

        if model_type == 'G':
            model = ModelFactory._create_g3d_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'ufdstvitnet':
            model = ModelFactory._create_unetfdstvit_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'U':
            model = ModelFactory._create_unet3d_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'uvit':
            model = ModelFactory._create_unet3dvit_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'gvit':
            model = ModelFactory._create_g3dvit_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'uattn':
            model = ModelFactory._create_unet3dattention_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'gattn':
            model = ModelFactory._create_g3dattention_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'udwtattn':
            model = ModelFactory._create_unet3ddwtattn_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'gwavevit':
            model = ModelFactory._create_g3dwavevit_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'shearmednet':
            model = ModelFactory._create_fdstunet3d_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'shearmedfdstvitnet':
            model = ModelFactory._create_shearmedfdstvit_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'cvshearmednet':
            model = ModelFactory._create_fdstcvaeunet3d_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'cvshearmedwavevitnet':
            model = ModelFactory._create_cvshearmedwavevitnet_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        # elif model_type == 'fdstcvaeunetfdstvitfdstcvaeunetfdstvit':
        #     model = ModelFactory._create_fdstcvaeunetfdstvit_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'glifting':
            model = ModelFactory._create_g3dlifting_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'gminimal':
            model = ModelFactory._create_g3dminimal_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'cvunet':
            model = ModelFactory._create_cvunet_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'cvuwavevitnet':
            model = ModelFactory._create_cvuwavevitnet_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'cvshearmedfdstvitnet':
            model = ModelFactory._create_cvshearmedfdstvitnet_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'cvufdstvitnet':
            model = ModelFactory._create_cvufdstvitnet_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'uwavevitnet':
            model = ModelFactory._create_unetwavevit_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'uiirwavevitnet':
            model = ModelFactory._create_unetwavevitediir_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        elif model_type == 'uiir':
            model = ModelFactory._create_unet3diir_model(input_size, config, resolved_n_classes, resolved_loss, class_weights)
        elif model_type == 'giirwavevit':
            model = ModelFactory._create_g3diirwavevit_model(input_size, config, resolved_n_classes, resolved_loss, class_weights, wave)
        else:
            # This should never happen due to the check above, but kept for safety
            raise ValueError(f"No creation method for model type: {model_type}")

        # Build/compile as requested
        if build:
            model.build(input_shape=(None, input_size, input_size, input_size, 1))
        if compile:
            # Uniform compile path: use provided optimizer or default Adam(1e-4)
            opt = optimizer or tf.keras.optimizers.Adam(1e-4)
            model.compile(optimizer=opt, jit_compile=jit_compile)
        if show_summary:
            try:
                model.summary()
            except Exception:
                pass
        return model

    @staticmethod
    def _create_g3d_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """
        Create G3D model with specified configuration.

        Args:
            input_size: Input dimension size
            config: Model configuration dictionary

        Returns:
            Compiled G3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        G = Gφψ3D(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            bottleneck_fn=get_bottleneck_fn(config['bottleneck_fn']),
            output_kernel_regularizer=None,
            residual=False,
            Ψ=(wave or 'haar')
        )

        return ModelFactory._finalize_model(G, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_unet3d_model(input_size, config, n_classes, loss, class_weights):
        """
        Create UNet3D model with specified configuration.

        Args:
            input_size: Input dimension size
            config: Model configuration dictionary

        Returns:
            Compiled UNet3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        if UNet3D is None:
            raise ValueError("UNet3D is not available in the current MEDNet package")

        unet_model = UNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            bottleneck_fn=get_bottleneck_fn(config['bottleneck_fn']),
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False
        )

        return ModelFactory._finalize_model(unet_model, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_unetfdstvit_model(input_size, config, n_classes, loss, class_weights):
        print(f"Creating {config['name']} model (input_size={input_size})...")
        try:
            from MEDNet.UFDSTViTNet3D import UNetFDSTViT3D
        except Exception as _e:
            raise ImportError("UNetFDSTViT3D is not available in MEDNet")
        import os as _os
        def _get_int(name, default):
            try:
                return int(_os.environ.get(name, default))
            except Exception:
                return default
        J = _get_int('MEDNET_FDST_J_BOTT', _get_int('MEDNET_FDST_J', 2))
        L1 = _get_int('MEDNET_FDST_L1_BOTT', _get_int('MEDNET_FDST_L1', 4))
        embed_dim = _get_int('MEDNET_VIT_EMBED', 64)
        heads = _get_int('MEDNET_VIT_HEADS', 2)
        inner = UNetFDSTViT3D(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            embed_dim=embed_dim,
            heads=heads,
            J=J,
            L1=L1,
            tokenizer="conv1d",
            token_conv_filters=None,
            gain_scale=1.0,
            use_feat_mixer=False,
            dropout=0.0,
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False,
        )
        try:
            inner.build((None, input_size, input_size, input_size, 1))
        except Exception:
            pass
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_unetwavevit_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """Create UNetWaveViT3D model (deterministic WaveletViT bottleneck) and wrap with common compile/loss handling."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if UNetWaveViT3D is None:
            raise ImportError("UNetWaveViT3D is not available in MEDNet")

        import os as _os
        def _get_int(name, default):
            try:
                return int(_os.environ.get(name, default))
            except Exception:
                return default
        # Wavelet-ViT sizing knobs (keep minimal and safe)
        wvit_embed   = _get_int('MEDNET_WVIT_EMBED', 64)
        wvit_heads   = _get_int('MEDNET_WVIT_HEADS', 2)
        wvit_keydim  = _get_int('MEDNET_WVIT_KEYDIM', 16)
        wvit_levels  = _get_int('MEDNET_WVIT_LEVELS', 1)

        inner = UNetWaveViT3D(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            embed_dim=wvit_embed,
            levels=wvit_levels,
            wave=(wave or 'haar'),
            heads=wvit_heads,
            key_dim=wvit_keydim,
            strength=0.25,
            level_decay=1.0,
            use_orient_mix=True,
            lp_global_if_small=True,
            max_tokens_global=8192,
            qkv_bias=True,
            dwt_compute_dtype='float32',
            prebuild_banks=True,
        )
        try:
            inner.build((None, input_size, input_size, input_size, 1))
        except Exception:
            pass
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_unetwavevitediir_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """Create UIIRDFIIWaveViTNet3D model (IIR conv blocks + deterministic WaveletViT bottleneck)."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if UIIRDFIIWaveViTNet3D is None:
            raise ImportError("UIIRDFIIWaveViTNet3D is not available in MEDNet")

        import os as _os
        def _get_int(name, default):
            try:
                return int(_os.environ.get(name, default))
            except Exception:
                return default
        # Wavelet-ViT sizing knobs (mirror uwavevitnet defaults)
        wvit_embed   = _get_int('MEDNET_WVIT_EMBED', 64)
        wvit_heads   = _get_int('MEDNET_WVIT_HEADS', 2)
        wvit_keydim  = _get_int('MEDNET_WVIT_KEYDIM', 8)
        wvit_levels  = _get_int('MEDNET_WVIT_LEVELS', 1)

        inner = UIIRDFIIWaveViTNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            embed_dim=wvit_embed,
            levels=wvit_levels,
            wave=(wave or 'haar'),
            heads=wvit_heads,
            key_dim=wvit_keydim,
            strength=0.25,
            level_decay=1.0,
            use_orient_mix=True,
            lp_global_if_small=True,
            max_tokens_global=2048,
            qkv_bias=True,
            dwt_compute_dtype='float32',
            prebuild_banks=True,
            iir_kwargs=None,
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False,
        )
        try:
            inner.build((None, input_size, input_size, input_size, 1))
        except Exception:
            pass
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_unet3dattention_model(input_size, config, n_classes, loss, class_weights):
        """
        Create UNet3Dattention model with specified configuration.

        Returns:
            Compiled UNet3Dattention wrapped G3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        unet_attn = UNet3Dattention(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            bottleneck_fn=get_bottleneck_fn(config.get('bottleneck_fn', 'vanilla_bottleneck')),
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False
        )

        return ModelFactory._finalize_model(unet_attn, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_unet3dvit_model(input_size, config, n_classes, loss, class_weights):
        """
        Create UNet3DViT model with specified configuration.

        Args:
            input_size: Input dimension size
            config: Model configuration dictionary

        Returns:
            Compiled UNet3DViT wrapped G3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        if UNet3DViT is None:
            raise ValueError("UNet3DViT is not available in the current MEDNet package")

        uvit = UNet3DViT(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False
        )

        return ModelFactory._finalize_model(uvit, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_g3dvit_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """
        Create G3DViT model with specified configuration.

        Returns:
            Compiled G3DViT wrapped G3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        if Gφψ3D_ViT is None:
            raise ValueError("G3DViT is not available in the current MEDNet package")

        gvit = Gφψ3D_ViT(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False,
            Ψ=(wave or 'haar')
        )

        return ModelFactory._finalize_model(gvit, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_g3dattention_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """
        Create G3Dattention model with specified configuration.

        Returns:
            Compiled G3Dattention wrapped G3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        gattn = G3Dattention(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            bottleneck_fn=get_bottleneck_fn(config.get('bottleneck_fn', 'vanilla_bottleneck')),
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False,
            Ψ=(wave or 'haar')
        )

        return ModelFactory._finalize_model(gattn, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_g3dwavevit_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """
        Create G3DwaveViT model with specified configuration.

        Returns:
            Compiled G3DwaveViT wrapped G3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        if Gφψ3D_waveViT is None:
            raise ValueError("G3DwaveViT is not available in the current MEDNet package")

        gwavevit = Gφψ3D_waveViT(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False,
            Ψ=(wave or 'haar')
        )

        return ModelFactory._finalize_model(gwavevit, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_fdstunet3d_model(input_size, config, n_classes, loss, class_weights):
        """
        Create FDSTUNet3D model with specified configuration.

        Returns:
            Compiled FDSTUNet3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        if FDSTUNet3D is None:
            raise ValueError("FDSTUNet3D is not available in the current MEDNet package")

        fdst_model = FDSTUNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            n_classes=n_classes,
            J=2,
            L1=4,
            base_filters=16,
            cone_mode='soft',
            rotation_mode='union',
            target_redundancy=2.0,
            feature_dropout=0.0,
            normalize_per_channel=False,
            final_activation=None,
            complex_repr='magnitude'
        )

        return ModelFactory._finalize_model(fdst_model, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_shearmedfdstvit_model(input_size, config, n_classes, loss, class_weights):
        """Create ShearMEDFDSTViTNet3D model and wrap with common compile/loss handling."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if ShearMEDFDSTViTNet3D is None:
            raise ImportError("ShearMEDFDSTViTNet3D is not available in MEDNet")

        import os as _os
        # Encoder FDST pyramid
        _J = int(_os.environ.get('MEDNET_FDST_J', '2'))
        _L1 = int(_os.environ.get('MEDNET_FDST_L1', '4'))
        # Bottleneck FDST pyramid (env-overridable; None lets factory fallback)
        def _maybe_int(name):
            v = _os.environ.get(name)
            try:
                return int(v) if v is not None else None
            except Exception:
                return None
        _Jb = _maybe_int('MEDNET_FDST_J_BOTT')
        _L1b = _maybe_int('MEDNET_FDST_L1_BOTT')
        # ViT sizing knobs
        def _get_int(name, default):
            try:
                return int(_os.environ.get(name, default))
            except Exception:
                return default
        _vit_embed = _get_int('MEDNET_VIT_EMBED', 64)
        _vit_heads = _get_int('MEDNET_VIT_HEADS', 2)

        inner = ShearMEDFDSTViTNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            n_classes=n_classes,
            J=_J,
            L1=_L1,
            base_filters=16,
            cone_mode='soft',
            rotation_mode='union',
            target_redundancy=2.0,
            normalize_per_channel=False,
            complex_repr='magnitude',
            embed_dim=_vit_embed,
            heads=_vit_heads,
            J_bott=_Jb,
            L1_bott=_L1b,
            tokenizer='conv1d',
            token_conv_filters=None,
            gain_scale=1.0,
            use_feat_mixer=False,
            dropout=0.0,
            final_activation=None,
        )
        try:
            inner.build((None, input_size, input_size, input_size, 1))
        except Exception:
            pass
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_fdstcvaeunet3d_model(input_size, config, n_classes, loss, class_weights):
        """Create FDSTCVAEUNet3D model with CVAE bottleneck and wrap with G3D."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if FDSTCVAEUNet3D is None:
            raise ValueError("FDSTCVAEUNet3D is not available in the current MEDNet package")
        inner = FDSTCVAEUNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            n_classes=n_classes,
            J=2,
            L1=4,
            base_filters=16,
            cone_mode='soft',
            rotation_mode='union',
            target_redundancy=2.0,
            feature_dropout=0.0,
            normalize_per_channel=False,
            final_activation=None,
            complex_repr='magnitude',
            latent_dim=128,
            fuse_skip=True,
        )
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_cvufdstvitnet_model(input_size, config, n_classes, loss, class_weights):
        """Create CVUFDSTViTNet3D model and wrap with common compile/loss handling."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if CVUFDSTViTNet3D is None:
            raise ImportError("CVUFDSTViTNet3D is not available in MEDNet")

        # Defaults: bottleneck J=1, L1=4, ViT embed 128, heads 4; allow env override
        import os as _os
        _Jb = int(_os.environ.get('MEDNET_FDST_J_BOTT', '1'))
        _L1b = int(_os.environ.get('MEDNET_FDST_L1_BOTT', '4'))
        # ViT sizing knobs
        def _get_int(name, default):
            try:
                return int(_os.environ.get(name, default))
            except Exception:
                return default
        _vit_embed = _get_int('MEDNET_VIT_EMBED', 128)
        _vit_heads = _get_int('MEDNET_VIT_HEADS', 4)
        # KL weight override
        def _get_float(name, default):
            try:
                return float(_os.environ.get(name, default))
            except Exception:
                return default
        _kl_scale = _get_float('MEDNET_VIT_KL_WEIGHT', 1e-4)

        inner = CVUFDSTViTNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            n_classes=n_classes,
            J_bott=_Jb,
            L1_bott=_L1b,
            vit_embed_dim=_vit_embed,
            vit_heads=_vit_heads,
            kl_scale=_kl_scale,
            # Use model defaults for optional features to keep factory clean
        )
        try:
            inner.build((None, input_size, input_size, input_size, 1))
        except Exception:
            pass
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_cvshearmedfdstvitnet_model(input_size, config, n_classes, loss, class_weights):
        """Create CVShearMEDFDSTViTNet3D model and wrap with compile/loss handling."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if FDSTCVAEUNetFDSTViT3D is None:
            raise ImportError("CVShearMEDFDSTViTNet3D is not available in MEDNet")

        import os as _os
        # Representation pyramid (skips)
        _J = int(_os.environ.get('MEDNET_FDST_J', '2'))
        _L1 = int(_os.environ.get('MEDNET_FDST_L1', '4'))
        # Bottleneck FDST settings
        _Jb = int(_os.environ.get('MEDNET_FDST_J_BOTT', '1'))
        _L1b = int(_os.environ.get('MEDNET_FDST_L1_BOTT', '4'))
        # ViT sizing knobs
        def _get_int(name, default):
            try:
                return int(_os.environ.get(name, default))
            except Exception:
                return default
        _vit_embed = _get_int('MEDNET_VIT_EMBED', 64)
        _vit_heads = _get_int('MEDNET_VIT_HEADS', 2)

        inner = FDSTCVAEUNetFDSTViT3D(
            in_shape=(input_size, input_size, input_size, 1),
            num_classes=n_classes,
            J=_J,
            L1=_L1,
            J_bott=_Jb,
            L1_bott=_L1b,
            embed_dim=_vit_embed,
            heads=_vit_heads,
            latent_channels=None,
            kl_weight=1e-4,
            use_feat_mixer=False,
            gain_scale=1.0,
            injection='residual',
            vae_mode='token',
            base_filters=16,
            feature_dropout=0.0,
            name=config['name'],
        )
        try:
            inner.build((None, input_size, input_size, input_size, 1))
        except Exception:
            pass
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_cvshearmedwavevitnet_model(input_size, config, n_classes, loss, class_weights):
        """Create CVShearMEDWaveViTNet3D model and wrap with G3D-like compile/loss handling."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if CVShearMEDWaveViTNet3D is None:
            raise ImportError("CVShearMEDWaveViTNet3D is not available.")
        inner = CVShearMEDWaveViTNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            n_classes=n_classes,
            J=2,
            L1=4,
            base_filters=16,
            cone_mode='soft',
            rotation_mode='union',
            target_redundancy=2.0,
            feature_dropout=0.0,
            normalize_per_channel=False,
            final_activation=None,
            complex_repr='magnitude',
            Ψ='haar',
            wvit_levels=1,
            vit_embed_dim=128,
            vit_heads=4,
            vit_depth=2,
            vit_key_dim=16,
            kl_scale=1.0,
        )
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    # def _create_fdstcvaeunetfdstvit_model(input_size, config, n_classes, loss, class_weights):
    #     """Removed: consolidated onto cvufdstvitnet (coefficient-domain FDSTViT) only."""
    #     raise NotImplementedError

    @staticmethod
    def _create_g3dlifting_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """
        Create G3DLifting model with specified configuration.

        Returns:
            Compiled G3DLifting wrapped G3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        if Gφψ3D_Lifting is None:
            raise ValueError("G3DLifting is not available in the current MEDNet package")

        glifting = Gφψ3D_Lifting(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            bottleneck_fn=get_bottleneck_fn(config['bottleneck_fn']),
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False,
            wavelet_refine=True,
            Ψ=(wave or 'haar')
        )

        return ModelFactory._finalize_model(glifting, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_g3dminimal_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """
        Create G3D_minimal model with specified configuration.

        Returns:
            Compiled minimal G3D wrapped model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        if Gφψ3D_Minimal is None:
            raise ValueError("G3D_minimal is not available in the current MEDNet package")

        inner = Gφψ3D_Minimal(
            Ψ=(wave or 'haar'),
            n_classes=n_classes,
            n_input_channels=1,
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            compile=False,
        )

        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    

    @staticmethod
    def _create_cvunet_model(input_size, config, n_classes, loss, class_weights):
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if CVAEUNet3D is None:
            raise ValueError("CVAEUNet3D is not available in the current MEDNet package")
        inner = CVAEUNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            n_classes=n_classes,
            config=configs[config['config_key']],
            latent_dim=512,
            compression_ratio=None,
            conv_fn=None,           # use CVAE defaults to avoid API mismatch
            pooling_fn=None,        # use internal default pooling
            upsampling_fn=None,     # use internal default upsampling
        )
        return inner

    @staticmethod
    def _create_cvuwavevitnet_model(input_size, config, n_classes, loss, class_weights, wave=None):
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if CVUWaveViTNet3D is None:
            raise ValueError("CVUWaveViTNet3D is not available in the current MEDNet package")
        # size knobs via env
        import os
        def _get_int(name, default):
            try:
                return int(os.environ.get(name, default))
            except Exception:
                return default
        vit_embed_dim = _get_int('MEDNET_WVIT_EMBED', 128)
        vit_depth     = _get_int('MEDNET_WVIT_DEPTH', 2)
        vit_heads     = _get_int('MEDNET_WVIT_HEADS', 4)
        vit_key_dim   = _get_int('MEDNET_WVIT_KEYDIM', 16)
        wvit_levels   = _get_int('MEDNET_WVIT_LEVELS', 1)

        inner = CVUWaveViTNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            Ψ=(wave or 'haar'),
            wvit_levels=wvit_levels,
            vit_embed_dim=vit_embed_dim,
            vit_heads=vit_heads,
            vit_depth=vit_depth,
            vit_key_dim=vit_key_dim,
        )
        # Wrap in G3D to unify compile, metrics (IoU/Dice), and logging with other models.
        # Inner model contributes KL via add_loss; G3D provides focal (or selected) loss and metrics.
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_unet3ddwtattn_model(input_size, config, n_classes, loss, class_weights):
        """
        Create UNet3Ddwtattn model with specified configuration.

        Returns:
            Compiled UNet3Ddwtattn wrapped G3D model
        """
        print(f"Creating {config['name']} model (input_size={input_size})...")

        unet_dwt_attn = UNet3Ddwtattn(
            input_shape=(input_size, input_size, input_size, 1),
            config=configs[config['config_key']],
            n_classes=n_classes,
            conv_fn=get_conv_fn(config['conv_fn']),
            pooling_fn=None,
            unpool_fn=None,
            bottleneck_fn=get_bottleneck_fn(config.get('bottleneck_fn', 'vanilla_bottleneck')),
            output_kernel_regularizer=None,
            one_hot_encode=True,
            residual=False,
            Ψ='haar'
        )

        return ModelFactory._finalize_model(unet_dwt_attn, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_unet3diir_model(input_size, config, n_classes, loss, class_weights):
        """Create UNet3DIIR model and wrap with G3D for loss/class weights."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if UIIRNet3D is None:
            raise ValueError("UNet3DIIR is not available in the current MEDNet package")
        # Optional channel override via env, e.g., MEDNET_CHANNELS="8,16,32,64"
        conf_tuple = configs[config['config_key']]
        ch_override = os.environ.get('MEDNET_CHANNELS')
        if ch_override:
            try:
                parsed = tuple(int(x.strip()) for x in ch_override.split(','))
                if len(parsed) == 4:
                    conf_tuple = parsed
                    print(f"[model_factory] Using MEDNET_CHANNELS override: {conf_tuple}")
            except Exception:
                pass

        inner = UIIRNet3D(
            input_shape=(input_size, input_size, input_size, 1),
            config=conf_tuple,
            n_classes=n_classes,
            one_hot_encode=True,
            residual=False,
        )
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _create_g3diirwavevit_model(input_size, config, n_classes, loss, class_weights, wave=None):
        """Create G3DIIRWaveViT model variant and wrap with G3D for training."""
        print(f"Creating {config['name']} model (input_size={input_size})...")
        if Gφψ3D_iirWaveViT is None:
            raise ValueError("G3DIIRWaveViT is not available in the current MEDNet package")
        # Env-driven light knobs to mitigate OOM without changing CLI
        # Channel widths override: e.g., MEDNET_CHANNELS="8,16,32,64"
        conf_tuple = configs[config['config_key']]
        ch_override = os.environ.get('MEDNET_CHANNELS')
        if ch_override:
            try:
                parsed = tuple(int(x.strip()) for x in ch_override.split(','))
                if len(parsed) == 4:
                    conf_tuple = parsed
                    print(f"[model_factory] Using MEDNET_CHANNELS override: {conf_tuple}")
            except Exception:
                pass

        # Wavelet-ViT size knobs
        def _get_int(name, default):
            try:
                return int(os.environ.get(name, default))
            except Exception:
                return default
        vit_embed_dim = _get_int('MEDNET_WVIT_EMBED', 128)
        vit_depth     = _get_int('MEDNET_WVIT_DEPTH', 2)
        vit_heads     = _get_int('MEDNET_WVIT_HEADS', 4)
        vit_key_dim   = _get_int('MEDNET_WVIT_KEYDIM', 16)
        wvit_levels   = _get_int('MEDNET_WVIT_LEVELS', 1)
        wvit_tokens   = _get_int('MEDNET_WVIT_TOKENS', 8192)

        inner = Gφψ3D_iirWaveViT(
            input_shape=(input_size, input_size, input_size, 1),
            config=conf_tuple,
            n_classes=n_classes,
            Ψ=(wave or 'haar'),
            one_hot_encode=True,
            residual=False,
            vit_embed_dim=vit_embed_dim,
            vit_depth=vit_depth,
            vit_heads=vit_heads,
            vit_key_dim=vit_key_dim,
            wvit_levels=wvit_levels,
            wvit_max_tokens_global=wvit_tokens,
        )
        return ModelFactory._finalize_model(inner, input_size, n_classes, loss, class_weights)

    @staticmethod
    def _finalize_model(inner_model, input_size, n_classes, loss, class_weights):
        """
        Finalize model by creating G3D wrapper, building, compiling, and summarizing.

        Args:
            inner_model: The inner model (Gφψ3D or UNet3D)
            input_size: Input dimension size for building
            config: Model configuration dictionary

        Returns:
            Finalized model
        """
        model = G3D(
            input_shape=(input_size, input_size, input_size, 1),
            model=inner_model,
            loss=loss,
            n_classes=n_classes,
            class_weights=class_weights,
        )
        return model

    @staticmethod
    def get_model_info(model_type):
        """
        Get information about a specific model type.

        Args:
            model_type: Model type identifier

        Returns:
            Dictionary with model information

        Raises:
            ValueError: If model_type is not supported
        """
        if model_type not in MODEL_CONFIGS:
            supported_models = list(MODEL_CONFIGS.keys())
            raise ValueError(f"Unknown model type: {model_type}. Supported: {supported_models}")

        return MODEL_CONFIGS[model_type].copy()

    @staticmethod
    def get_supported_models():
        """
        Get list of supported model types.

        Returns:
            List of supported model type identifiers
        """
        return list(MODEL_CONFIGS.keys())

def create_model(model_type, input_size=64, **kwargs):
    """
    Convenience function to create a model using the factory.

    This function provides backward compatibility and a simpler interface
    for creating models.

    Args:
        model_type: Model type identifier ('G' for G3D, 'U' for UNet3D)
        input_size: Input dimension size

    Returns:
        Keras model
    """
    return ModelFactory.create_model(model_type, input_size, **kwargs)

def get_supported_models():
    """
    Convenience function to get supported model types.

    Returns:
        List of supported model type identifiers
    """
    return ModelFactory.get_supported_models()

def get_model_info(model_type):
    """
    Convenience function to get model information.

    Args:
        model_type: Model type identifier

    Returns:
        Dictionary with model information
    """
    return ModelFactory.get_model_info(model_type)



from __future__ import annotations
from typing import Tuple, Optional
import numpy as np


def bbox_from_mask(mask3d: np.ndarray, margin: int = 8) -> Tuple[slice, slice, slice]:
    """Compute tight bbox around nonzero mask with margin; returns slices (z,y,x)."""
    assert mask3d.ndim == 3
    nz = np.nonzero(mask3d)
    if len(nz[0]) == 0:
        # Empty mask -> full volume
        D, H, W = mask3d.shape
        return (slice(0, D), slice(0, H), slice(0, W))
    zmin, zmax = np.min(nz[0]), np.max(nz[0])
    ymin, ymax = np.min(nz[1]), np.max(nz[1])
    xmin, xmax = np.min(nz[2]), np.max(nz[2])
    D, H, W = mask3d.shape
    z0 = max(0, zmin - margin); z1 = min(D, zmax + 1 + margin)
    y0 = max(0, ymin - margin); y1 = min(H, ymax + 1 + margin)
    x0 = max(0, xmin - margin); x1 = min(W, xmax + 1 + margin)
    return (slice(z0, z1), slice(y0, y1), slice(x0, x1))


def crop3d(vol: np.ndarray, bbox: Tuple[slice, slice, slice]) -> np.ndarray:
    return vol[bbox[0], bbox[1], bbox[2], ...]


def pad_to_multiple_3d(vol: np.ndarray, multiple: int) -> Tuple[np.ndarray, Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int]]]:
    """Zero-pad a 3D or 4D volume so D,H,W are multiples of 'multiple'. Returns padded vol and pad sizes per axis."""
    is4 = (vol.ndim == 4)
    if is4:
        D, H, W, C = vol.shape
    else:
        D, H, W = vol.shape
    def _pad_for(n):
        rem = n % multiple
        add = (multiple - rem) % multiple
        return (0, add)
    pd = _pad_for(D); ph = _pad_for(H); pw = _pad_for(W)
    pad_cfg = (pd, ph, pw, (0, 0)) if is4 else (pd, ph, pw)
    out = np.pad(vol, pad_cfg, mode="constant")
    return out, (pd, ph, pw)


def unpad3d(vol: np.ndarray, pads: Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int]]) -> np.ndarray:
    (pd, ph, pw) = pads
    D2 = vol.shape[0] - pd[1]
    H2 = vol.shape[1] - ph[1]
    W2 = vol.shape[2] - pw[1]
    return vol[:D2, :H2, :W2, ...]


def paste_back_3d(full_shape: Tuple[int, int, int], crop_labels: np.ndarray, bbox: Tuple[slice, slice, slice], fill: int = 0) -> np.ndarray:
    """Paste cropped labels back into a full-sized array."""
    out = np.full(full_shape, fill_value=fill, dtype=crop_labels.dtype)
    out[bbox[0], bbox[1], bbox[2]] = crop_labels
    return out


def pad_to_shape_cube_3d(vol: np.ndarray, size: int = 256) -> Tuple[np.ndarray, Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int]]]:
    """Zero-pad a 3D or 4D volume so D=H=W=size using symmetric padding (center-pad).

    Returns (padded_volume, pads), where pads are (pd, ph, pw) tuples of (before, after).
    """
    is4 = (vol.ndim == 4)
    if is4:
        D, H, W, C = vol.shape
    else:
        D, H, W = vol.shape
    def _sym_pad(n):
        add = max(0, size - n)
        before = add // 2
        after = add - before
        return (before, after)
    pd = _sym_pad(D); ph = _sym_pad(H); pw = _sym_pad(W)
    pad_cfg = (pd, ph, pw, (0, 0)) if is4 else (pd, ph, pw)
    out = np.pad(vol, pad_cfg, mode="constant")
    return out, (pd, ph, pw)



from __future__ import annotations
import os
from datetime import datetime


def make_timestamped_output_dir(base_dir: str) -> str:
    """Create and return a timestamped subdirectory under base_dir.

    Format: MonDay_output_HHMMSS (e.g., Nov9_output_142530)
    """
    now = datetime.now()
    day_mon = f"{now.strftime('%b')}{now.day}"
    hhmmss = now.strftime("%H%M%S")
    out_dir = os.path.join(base_dir, f"{day_mon}_output_{hhmmss}")
    os.makedirs(out_dir, exist_ok=True)
    return out_dir




import os
import tensorflow as tf
from tensorflow.keras import layers as L

from MEDNet.UNet3D import UNet3D
from MEDNet.attentions.WaveletViTND import (
    WaveletTokenizer3D,
    WaveletViTBlock3D,
    WaveletAssembler3D,
)

__all__ = [
    "wavevit_bottleneck_factory",
    "UNetWaveViT3D",
]


def _env_int(name: str, default: int) -> int:
    try:
        v = os.getenv(name)
        return int(v) if v is not None else int(default)
    except Exception:
        return int(default)


def _env_float(name: str, default: float) -> float:
    try:
        v = os.getenv(name)
        return float(v) if v is not None else float(default)
    except Exception:
        return float(default)


def wavevit_bottleneck_factory(
    embed_dim: int = 64,
    levels: int = 1,
    wave: str = "haar",
    *,
    heads: int = 2,
    key_dim: int = 16,
    strength: float = 0.25,
    level_decay: float = 1.0,
    use_orient_mix: bool = True,
    lp_global_if_small: bool = True,
    max_tokens_global: int = 8192,
    qkv_bias: bool = True,
    dwt_compute_dtype: str = "float32",
    prebuild_banks: bool = True,
):
    """Create a deterministic Wavelet-ViT 3D bottleneck callable for UNet3D.

    Returns a function bottleneck(x) -> features using:
      WaveletTokenizer3D -> WaveletViTBlock3D -> WaveletAssembler3D(mode='feature')
    Produces a spatial feature map with `embed_dim` channels.
    """
    E = int(embed_dim)
    Lvl = int(levels)
    H = int(heads)
    KD = int(key_dim)

    # Build layers once, reuse per-call (functional graph friendly)
    tok = WaveletTokenizer3D(levels=Lvl, wave=wave, embed_dim=E,
                             dwt_compute_dtype=dwt_compute_dtype, prebuild_banks=prebuild_banks)
    blk = WaveletViTBlock3D(heads=H, key_dim=KD, strength=float(strength),
                             level_decay=float(level_decay), use_orient_mix=bool(use_orient_mix),
                             lp_global_if_small=bool(lp_global_if_small), max_tokens_global=int(max_tokens_global),
                             qkv_bias=bool(qkv_bias))
    asm = WaveletAssembler3D(levels=Lvl, wave=wave, mode='feature',
                              dwt_compute_dtype=dwt_compute_dtype, prebuild_banks=prebuild_banks)

    def bottleneck(x: tf.Tensor) -> tf.Tensor:
        bands = tok(x)
        bands = blk(bands)
        y = asm(bands)
        # Ensure dtype consistency with upstream
        if y.dtype != x.dtype:
            y = tf.cast(y, x.dtype)
        return y

    return bottleneck


def UNetWaveViT3D(
    input_shape=(32, 32, 32, 1),
    config=(16, 32, 64, 128),
    n_classes=4,
    *,
    embed_dim: int | None = None,
    levels: int = 1,
    wave: str = "haar",
    heads: int = 2,
    key_dim: int = 16,
    strength: float = 0.25,
    level_decay: float = 1.0,
    use_orient_mix: bool = True,
    lp_global_if_small: bool = True,
    max_tokens_global: int = 8192,
    qkv_bias: bool = True,
    dwt_compute_dtype: str = "float32",
    prebuild_banks: bool = True,
    output_kernel_regularizer=None,
    one_hot_encode=True,
    residual=False,
):
    """UNet3D with a deterministic Wavelet-ViT (3D) bottleneck (no VAE).

    The bottleneck operates in the DWT band domain and returns a spatial
    feature with `embed_dim` channels (defaulting to config[-1]).
    """
    E = int(embed_dim) if embed_dim is not None else int(config[-1])
    bottleneck = wavevit_bottleneck_factory(
        embed_dim=E,
        levels=levels,
        wave=wave,
        heads=heads,
        key_dim=key_dim,
        strength=strength,
        level_decay=level_decay,
        use_orient_mix=use_orient_mix,
        lp_global_if_small=lp_global_if_small,
        max_tokens_global=max_tokens_global,
        qkv_bias=qkv_bias,
        dwt_compute_dtype=dwt_compute_dtype,
        prebuild_banks=prebuild_banks,
    )
    return UNet3D(
        input_shape=input_shape,
        config=config,
        n_classes=n_classes,
        bottleneck_fn=bottleneck,
        output_kernel_regularizer=output_kernel_regularizer,
        one_hot_encode=one_hot_encode,
        residual=residual,
    )


if __name__ == "__main__":
    tf.get_logger().setLevel('ERROR')
    tf.random.set_seed(0)
    n = 32
    model = UNetWaveViT3D(
        input_shape=(n, n, n, 1),
        config=(16, 32, 64, 128),
        n_classes=4,
        embed_dim=64,
        levels=1,
        wave='haar',
        heads=2,
        key_dim=16,
        strength=0.5,
        level_decay=1.0,
    )
    x = tf.random.normal((1, n, n, n, 1))
    y = model(x)
    model.summary()
    print("out:", y.shape)



""" MRAViT: Multiresolution D-dimensional Attentions for Vision
    Copyright (C) 2025 Kishore Kumar Tarafdar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

WaveletViTND.py
Native Wavelet-ViT (1D/2D/3D): DWT tokenizer -> band-wise axial attention (+ optional global on deepest LP)
-> light orientation mixing -> IDWT assembler (tokens or feature).
Keras-graph safe (tensors-only I/O), eager filter-bank prebuild, build-time state creation.
"""
import tensorflow as tf
from tensorflow import keras

# ---------- Robust DWT/IDWT imports ----------
HAVE_1D = False
HAVE_3D = False
try:
    from TFDWT.DWT2DFB import DWT2D, IDWT2D
except Exception:
    from DWT2DFB import DWT2D, IDWT2D

try:
    from TFDWT.DWT3DFB import DWT3D, IDWT3D
    HAVE_3D = True
except Exception:
    try:
        from DWT3DFB import DWT3D, IDWT3D
        HAVE_3D = True
    except Exception:
        HAVE_3D = False

try:
    from TFDWT.DWT1DFB import DWT1D, IDWT1D
    HAVE_1D = True
except Exception:
    try:
        from DWT1DFB import DWT1D, IDWT1D
        HAVE_1D = True
    except Exception:
        HAVE_1D = False


# ---------- Attention blocks ----------
@tf.keras.utils.register_keras_serializable(package="attn")
class GlobalMHSAND(keras.layers.Layer):
    def __init__(self, heads=2, key_dim=16, norm_eps=1e-5, qkv_bias=True, **kw):
        super().__init__(**kw)
        self.heads = int(heads); self.key_dim = int(key_dim)
        self.norm = keras.layers.LayerNormalization(epsilon=norm_eps)
        self.qkv_bias = bool(qkv_bias); self.Conv = None

    def build(self, x_shape):
        dims = len(x_shape) - 2; C = x_shape[-1]
        Conv = {1: keras.layers.Conv1D, 2: keras.layers.Conv2D, 3: keras.layers.Conv3D}[dims]
        self.Conv = Conv
        self.to_q = Conv(self.heads * self.key_dim, 1, use_bias=self.qkv_bias)
        self.to_k = Conv(self.heads * self.key_dim, 1, use_bias=self.qkv_bias)
        self.to_v = Conv(self.heads * self.key_dim, 1, use_bias=self.qkv_bias)
        self.proj = Conv(C, 1, use_bias=self.qkv_bias)

    def call(self, x):
        X = self.norm(x)
        Q = self.to_q(X); K = self.to_k(X); V = self.to_v(X)
        B = tf.shape(x)[0]; spatial = tf.reduce_prod(tf.shape(x)[1:-1])
        h, d = self.heads, self.key_dim

        def pack(t):
            t = tf.reshape(t, [B, spatial, h, d])
            return tf.transpose(t, [0, 2, 1, 3])  # [B,h,T,d]

        Q, K, V = map(pack, (Q, K, V))
        scale = tf.cast(self.key_dim, x.dtype) ** -0.5
        A = tf.nn.softmax(tf.matmul(Q, K, transpose_b=True) * scale, axis=-1)
        Y = tf.matmul(A, V)
        Y = tf.transpose(Y, [0, 2, 1, 3])
        Y = tf.reshape(Y, tf.concat([[B], tf.shape(x)[1:-1], [h * d]], axis=0))
        return self.proj(Y)


@tf.keras.utils.register_keras_serializable(package="attn")
class AxialMHSAND(keras.layers.Layer):
    def __init__(self, heads=2, key_dim=16, norm_eps=1e-5, qkv_bias=True, **kw):
        super().__init__(**kw)
        self.heads = int(heads); self.key_dim = int(key_dim)
        self.norm = keras.layers.LayerNormalization(epsilon=norm_eps)
        self.qkv_bias = bool(qkv_bias); self.Conv = None; self.dims = None

    def build(self, x_shape):
        self.dims = len(x_shape) - 2; C = x_shape[-1]
        self.Conv = {1: keras.layers.Conv1D, 2: keras.layers.Conv2D, 3: keras.layers.Conv3D}[self.dims]
        self.to_q = self.Conv(self.heads * self.key_dim, 1, use_bias=self.qkv_bias)
        self.to_k = self.Conv(self.heads * self.key_dim, 1, use_bias=self.qkv_bias)
        self.to_v = self.Conv(self.heads * self.key_dim, 1, use_bias=self.qkv_bias)
        self.proj = self.Conv(C, 1, use_bias=self.qkv_bias)

    @staticmethod
    def _axis_perm(dims, axis):
        spatial = list(range(1, 1 + dims)); axis_idx = 1 + axis
        others = [i for i in spatial if i != axis_idx]
        perm = [0, axis_idx] + others + [dims + 1]
        inv = [0] * (dims + 2)
        for i, p in enumerate(perm): inv[p] = i
        return perm, inv

    def _attend_along_axis(self, x, axis):
        B = tf.shape(x)[0]; X = self.norm(x)
        Q = self.to_q(X); K = self.to_k(X); V = self.to_v(X)
        h, d = self.heads, self.key_dim
        scale = tf.cast(d, x.dtype) ** -0.5

        perm, inv = self._axis_perm(self.dims, axis)
        Qp = tf.transpose(Q, perm); Kp = tf.transpose(K, perm); Vp = tf.transpose(V, perm)
        shp = tf.shape(Qp); T = shp[1]
        O = tf.reduce_prod(shp[2:-1]) if self.dims > 1 else 1

        def pack(t):
            t = tf.reshape(t, tf.concat([[B, T, O], [h * d]], axis=0))
            t = tf.reshape(t, [B * O, T, h, d])
            return tf.transpose(t, [0, 2, 1, 3])

        Qb = pack(Qp); Kb = pack(Kp); Vb = pack(Vp)
        Ab = tf.nn.softmax(tf.matmul(Qb, Kb, transpose_b=True) * scale, axis=-1)
        Yb = tf.matmul(Ab, Vb)
        Yb = tf.transpose(Yb, [0, 2, 1, 3])
        Yb = tf.reshape(Yb, [B, O, T, h * d])
        Yp = tf.reshape(Yb, tf.concat([[B], [T], shp[2:-1], [h * d]], axis=0)) if self.dims > 1 else tf.reshape(Yb, [B, T, h * d])
        Y = tf.transpose(Yp, inv)
        return self.proj(Y)

    def call(self, x):
        y = x
        for axis in range(self.dims):
            y = self._attend_along_axis(y, axis)
        return y


@tf.keras.utils.register_keras_serializable(package="attn")
class OrientationMix(keras.layers.Layer):
    """Mix across G=2^D-1 HP groups (channel-wise)."""
    def __init__(self, groups, eps=1e-3, **kw):
        super().__init__(**kw)
        self.G = int(groups); self.C = None; self.eps = float(eps); self.W=None

    def build(self, x_shape):
        self.C = x_shape[-1] // self.G
        import numpy as np
        G = self.G
        base = np.eye(G, dtype='float32'); off = np.ones((G, G), dtype='float32') - base
        init_val = base + self.eps * off
        self.W = self.add_weight(name="W", shape=(G, G),
                                 initializer=keras.initializers.Constant(init_val),
                                 trainable=True)

    def call(self, x):
        shp = tf.shape(x); G, C = self.G, self.C
        y = tf.reshape(x, tf.concat([shp[:-1], [G, C]], axis=0))
        y = tf.einsum('gd,...dc->...gc', self.W, y)
        return tf.reshape(y, tf.concat([shp[:-1], [G * C]], axis=0))


# ---------- Tokenizer ----------
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletTokenizerND(keras.layers.Layer):
    """
    DWT tokenization into levels × subbands with per-band 1×1 projections to embed_dim.
    Returns `bands`: list[level] of {'lp': Tensor[B,*Sℓ,E], 'hp': List[Tensor[B,*Sℓ,E]]}.
    """
    def __init__(self, dims=2, levels=2, wave='sym8', embed_dim=256, norm=True,
                 pad_to_pow2=True, dwt_compute_dtype='float32', prebuild_banks=True, name=None, **kw):
        super().__init__(name=name, **kw)
        assert dims in (1,2,3)
        self.D = int(dims); self.levels = int(levels); self.wave = wave
        self.embed_dim = int(embed_dim); self.norm = bool(norm)
        self.pad_to_pow2 = bool(pad_to_pow2); self.dwt_compute_dtype = dwt_compute_dtype
        self.prebuild_banks = bool(prebuild_banks)

        if self.D==1:
            if not HAVE_1D: raise ImportError("DWT1D/IDWT1D not available")
            self.DWT = DWT1D
        elif self.D==2:
            self.DWT = DWT2D
        else:
            if not HAVE_3D: raise ImportError("DWT3D/IDWT3D not available")
            self.DWT = DWT3D

        self.total_bands = 2**self.D; self.groups = self.total_bands - 1
        # Force DWT layers to compute in float32 to avoid bf16 dtype issues in einsum
        self.dwts = [self.DWT(wave=self.wave, clean=True, dtype='float32') for _ in range(self.levels)]
        self.norm_layer = keras.layers.LayerNormalization(epsilon=1e-5) if self.norm else None
        Conv = {1: keras.layers.Conv1D, 2: keras.layers.Conv2D, 3: keras.layers.Conv3D}[self.D]
        self.proj_lp = Conv(self.embed_dim, 1, use_bias=True)
        self.proj_hp = Conv(self.embed_dim, 1, use_bias=True)

    def _pad_pow2(self, x):
        if not self.pad_to_pow2:
            return x
        shp = tf.shape(x)
        spatial = [shp[1+i] for i in range(self.D)]
        div = tf.constant(2**self.levels, tf.int32)
        adds = [tf.math.floormod(div - tf.math.floormod(n, div), div) for n in spatial]
        paddings = tf.stack([[0,0]] + [[0, a] for a in adds] + [[0,0]], axis=0)
        return tf.pad(x, paddings)

    def _prebuild_banks_eager(self, x_shape):
        with tf.init_scope():
            spatial = [int(s) if (s is not None and s!=-1) else 16 for s in x_shape[1:-1]]
            C = int(x_shape[-1]) if x_shape[-1] is not None else 1
            div = 2**self.levels
            spatial = [s - (s % div) if s % div != 0 else s for s in spatial]
            dummy = tf.zeros([1]+spatial+[C], tf.float32)
            cur = dummy
            for _ in range(self.levels):
                cur = self.dwts[_](cur)[..., :C]

    def build(self, x_shape):
        if self.prebuild_banks: self._prebuild_banks_eager(x_shape)
        super().build(x_shape)

    def call(self, x):
        if self.norm_layer is not None: x = self.norm_layer(x)
        x = self._pad_pow2(x)

        C_in = tf.shape(x)[-1]
        cur = x
        bands = []
        tgt_dtype = tf.as_dtype(self.dwt_compute_dtype)

        for _ in range(self.levels):
            cur32 = tf.cast(cur, tgt_dtype) if cur.dtype != tgt_dtype else cur
            w = self.dwts[_](cur32)             # [B,*Sℓ, total_bands*C_in]
            w = tf.cast(w, x.dtype)
            lp = w[..., :C_in]                  # LL
            hp_concat = w[..., C_in:]           # HPs concatenated
            hp_list = tf.split(hp_concat, num_or_size_splits=self.groups, axis=-1)
            lp_e = self.proj_lp(lp)
            hp_e = [self.proj_hp(h) for h in hp_list]
            bands.append({'lp': lp_e, 'hp': hp_e})
            cur = lp
        return bands


# ---------- Transformer block over bands ----------
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletViTBlockND(keras.layers.Layer):
    """
    Band-wise axial attention + optional global on deepest LP, OrientationMix, per-level decay, delta residual.
    Input/Output: same bands structure as tokenizer.
    """
    def __init__(self, dims=2, heads=2, key_dim=16, strength=0.25, level_decay=0.6,
                 use_orient_mix=True, lp_global_if_small=True, max_tokens_global=8192, qkv_bias=True,
                 **kw):
        super().__init__(**kw)
        assert dims in (1,2,3)
        self.D = int(dims); self.heads=int(heads); self.key_dim=int(key_dim)
        self.strength = float(strength); self.level_decay = float(level_decay)
        self.use_orient_mix = bool(use_orient_mix)
        self.lp_global_if_small = bool(lp_global_if_small); self.max_tokens_global = int(max_tokens_global)
        self.qkv_bias = bool(qkv_bias)

        self.hp_attn = AxialMHSAND(self.heads, self.key_dim, qkv_bias=self.qkv_bias)
        self.lp_axial = AxialMHSAND(self.heads, self.key_dim, qkv_bias=self.qkv_bias)
        self.lp_global = GlobalMHSAND(self.heads, self.key_dim, qkv_bias=self.qkv_bias)
        self.mixers = None
        self.alpha_hp = None; self.alpha_lp = None
        self.groups = 2**self.D - 1
        self.levels_ = None

    def build(self, bands_shape):
        try:
            self.levels_ = len(bands_shape)
        except Exception:
            self.levels_ = 1
        self.alpha_hp = self.add_weight(name="alpha_hp", shape=(),
                                        initializer=keras.initializers.Constant(0.1), trainable=True)
        self.alpha_lp = self.add_weight(name="alpha_lp", shape=(),
                                        initializer=keras.initializers.Constant(0.1), trainable=True)
        mix_eps = 0.02 * float(self.strength)
        self.mixers = [OrientationMix(self.groups, eps=mix_eps) for _ in range(self.levels_)]
        super().build(bands_shape)

    def _tokens(self, spatial_shape):
        return tf.cast(tf.reduce_prod(spatial_shape), tf.int32)

    def call(self, bands):
        out = []
        L = len(bands)
        for l, d in enumerate(bands):
            lp = d['lp']; hp_list = d['hp']
            level_scale = tf.cast(self.level_decay ** (L-1 - l), lp.dtype)

            hp_proc = [h + (self.alpha_hp * level_scale) * self.hp_attn(h) for h in hp_list]

            if self.use_orient_mix and len(hp_proc) > 1:
                cat = tf.concat(hp_proc, axis=-1)
                cat = self.mixers[l](cat)
                hp_proc = tf.split(cat, num_or_size_splits=len(hp_proc), axis=-1)

            spatial = tf.shape(lp)[1:-1]
            is_last = (l == L - 1)
            def do_axial():  return lp + (self.alpha_lp * level_scale) * self.lp_axial(lp)
            def do_global(): return lp + (self.alpha_lp * level_scale) * self.lp_global(lp)
            if self.lp_global_if_small and is_last:
                T = self._tokens(spatial)
                lp_tokens_small = tf.less_equal(T, tf.cast(self.max_tokens_global, tf.int32))
                lp = tf.cond(lp_tokens_small, do_global, do_axial)
            else:
                lp = do_axial()

            out.append({'lp': lp, 'hp': hp_proc})

        s = tf.cast(self.strength, out[0]['lp'].dtype)
        bands_out = []
        for din, dout in zip(bands, out):
            lp = din['lp'] + s * (dout['lp'] - din['lp'])
            hp = [h_in + s * (h_out - h_in) for h_in, h_out in zip(din['hp'], dout['hp'])]
            bands_out.append({'lp': lp, 'hp': hp})
        return bands_out


# ---------- Assembler ----------
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletAssemblerND(keras.layers.Layer):
    """
    mode='tokens'  -> pooled deepest LP: [B, E]
    mode='feature' -> IDWT cascade back to feature map: [B, *orig, E]
    Optional cropping can be added via a tensor `pads` if needed (omitted here for simplicity).
    """
    def __init__(self, dims=2, levels=2, wave='sym8', mode='tokens',
                 dwt_compute_dtype='float32', prebuild_banks=True, name=None, **kw):
        super().__init__(name=name, **kw)
        assert dims in (1,2,3); assert mode in ('tokens','feature')
        self.D=int(dims); self.levels=int(levels); self.wave=wave; self.mode=mode
        self.dwt_compute_dtype = dwt_compute_dtype; self.prebuild_banks = bool(prebuild_banks)

        if self.D==1:
            if not HAVE_1D: raise ImportError("IDWT1D not available")
            self.IDWT = IDWT1D
        elif self.D==2:
            self.IDWT = IDWT2D
        else:
            if not HAVE_3D: raise ImportError("IDWT3D not available")
            self.IDWT = IDWT3D

        # Force IDWT layers to compute in float32 to match DWT and avoid bf16 incompatibilities
        self.idwts = [self.IDWT(wave=self.wave, clean=True, dtype='float32') for _ in range(self.levels)]
        self.pool = None

    def build(self, bands_shape):
        if self.mode=='tokens':
            self.pool = keras.layers.GlobalAveragePooling1D() if self.D==1 else \
                        keras.layers.GlobalAveragePooling2D() if self.D==2 else \
                        keras.layers.GlobalAveragePooling3D()
        super().build(bands_shape)

    def _prebuild_banks_eager(self, bands):
        with tf.init_scope():
            spatial = []
            for s in bands[-1]['lp'].shape[1:-1]:
                spatial.append(int(s) if s is not None else 16)
            E = int(bands[-1]['lp'].shape[-1]) if bands[-1]['lp'].shape[-1] is not None else 32
            for l in reversed(range(self.levels)):
                dummy_low = tf.zeros([1]+spatial+[E], tf.float32)
                dummy_hp  = tf.zeros([1]+spatial+[(2**self.D - 1) * E], tf.float32)
                _ = self.idwts[l](tf.concat([dummy_low, dummy_hp], axis=-1))
                spatial = [s*2 for s in spatial]

    def call(self, bands, pads=None):
        if self.mode == 'tokens':
            deepest_lp = bands[-1]['lp']
            return self.pool(deepest_lp)
        if self.prebuild_banks: self._prebuild_banks_eager(bands)

        tgt_dtype = tf.as_dtype(self.dwt_compute_dtype)
        cur = None
        for rev_l, d in enumerate(reversed(bands)):
            low = d['lp']; hp_list = d['hp']
            hp = tf.concat(hp_list, axis=-1) if len(hp_list) > 0 else tf.zeros_like(low)[..., :0]
            merged = tf.concat([low, hp], axis=-1)
            merged32 = tf.cast(merged, tgt_dtype) if merged.dtype != tgt_dtype else merged
            cur = self.idwts[self.levels - 1 - rev_l](merged32)
            cur = tf.cast(cur, bands[0]['lp'].dtype)
        return cur


# ---------- Convenience wrappers ----------
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletTokenizer1D(WaveletTokenizerND):  
    def __init__(self, **kw): super().__init__(dims=1, **kw)
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletTokenizer2D(WaveletTokenizerND):  
    def __init__(self, **kw): super().__init__(dims=2, **kw)
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletTokenizer3D(WaveletTokenizerND):  
    def __init__(self, **kw): super().__init__(dims=3, **kw)

@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletViTBlock1D(WaveletViTBlockND):  
    def __init__(self, **kw): super().__init__(dims=1, **kw)
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletViTBlock2D(WaveletViTBlockND):  
    def __init__(self, **kw): super().__init__(dims=2, **kw)
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletViTBlock3D(WaveletViTBlockND):  
    def __init__(self, **kw): super().__init__(dims=3, **kw)

@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletAssembler1D(WaveletAssemblerND):  
    def __init__(self, **kw): super().__init__(dims=1, **kw)
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletAssembler2D(WaveletAssemblerND):  
    def __init__(self, **kw): super().__init__(dims=2, **kw)
@tf.keras.utils.register_keras_serializable(package="attn")
class WaveletAssembler3D(WaveletAssemblerND):  
    def __init__(self, **kw): super().__init__(dims=3, **kw)


# ---------- Test helpers ----------
def _rel_l2(a, b):
    num = tf.norm(tf.reshape(a - b, [tf.shape(a)[0], -1]), ord='euclidean', axis=1)
    den = tf.norm(tf.reshape(b,     [tf.shape(b)[0], -1]), ord='euclidean', axis=1) + 1e-12
    return tf.reduce_max(num / den)

def _bands_rel_l2(b1, b2):
    vals = []
    for d1, d2 in zip(b1, b2):
        vals.append(_rel_l2(d1['lp'], d2['lp']))
        for h1, h2 in zip(d1['hp'], d2['hp']):
            vals.append(_rel_l2(h1, h2))
    return float(tf.reduce_max(tf.stack(vals)))

def _bandspace_max_abs_delta(b_in, b_out):
    out = []
    for l, (d0, d1) in enumerate(zip(b_in, b_out)):
        lp = float(tf.reduce_max(tf.abs(d1['lp'] - d0['lp'])))
        c0 = tf.concat(d0['hp'], axis=-1) if d0['hp'] else tf.zeros_like(d0['lp'])[..., :0]
        c1 = tf.concat(d1['hp'], axis=-1) if d1['hp'] else tf.zeros_like(d1['lp'])[..., :0]
        hp = float(tf.reduce_max(tf.abs(c1 - c0)))
        out.append((lp, hp))
    return out

def _grad_report(model, x, y, keys=None):
    keys = keys or []
    with tf.GradientTape() as tape:
        yhat = model(x, training=True)
        loss = tf.reduce_mean(tf.square(yhat - y))
    grads = tape.gradient(loss, model.trainable_variables)
    total = sum(float(tf.norm(g)) for g in grads if g is not None)
    bucket = {k:0.0 for k in keys}
    for var, g in zip(model.trainable_variables, grads):
        if g is None: continue
        name = var.name
        for k in keys:
            if k in name:
                bucket[k] += float(tf.norm(g))
    print("Grad totals:", {k: round(v, 6) for k,v in bucket.items()}, "| total:", round(total, 6))
    return total, bucket


# ---------- Self-tests & demos ----------
# ------------------------- Self-tests & demos -------------------------
if __name__ == "__main__":
    try:
        has_gpu = len(tf.config.list_physical_devices('GPU')) > 0
        if has_gpu:
            tf.keras.mixed_precision.set_global_policy('mixed_float16')
            print("Mixed precision: enabled (GPU detected)")
        else:
            print("Mixed precision: skipped (no GPU)")
    except Exception:
        print("Mixed precision: skipped")

    tf.random.set_seed(0)
    print("Running WaveletViTND self-tests (PR + grad + serialization) ...")

    # ========================= 2D (tokens) LP-only (HP frozen to silence warnings) =========================
    print("\n=== 2D (tokens, LP-only) ===")
    B, N, C = 2, 128, 3
    x2 = tf.random.normal((B, N, N, C))

    # --- PR-2: DWT->IDWT transform round-trip sanity check (2D) ---
    try:
        dwt_rt = DWT2D(wave='sym8', clean=True)
        idwt_rt = IDWT2D(wave='sym8', clean=True)
        x2_rt_in = tf.cast(x2, tf.float32)  # compute in fp32 for crisp PR
        w_rt = dwt_rt(x2_rt_in)
        x2_rt = idwt_rt(w_rt)
        rel_rt = float(_rel_l2(x2_rt, x2_rt_in))
        print("2D | PR-2 (transform round-trip) rel-L2:", rel_rt)
    except Exception as e:
        print("2D | PR-2 check skipped:", e)

    tok2_lp = WaveletTokenizer2D(levels=2, wave='sym8', embed_dim=64)
    b2_in = tok2_lp(x2)
    blk2_lp_pr = WaveletViTBlock2D(heads=2, key_dim=16, strength=0.0)
    b2_out_pr = blk2_lp_pr(b2_in)
    print("2D | PR (strength=0) bands rel-L2:", _bands_rel_l2(b2_out_pr, b2_in))

    blk2_lp = WaveletViTBlock2D(heads=2, key_dim=16, strength=0.25)
    b2_out = blk2_lp(b2_in)
    print("2D | default strength bands rel-L2:", _bands_rel_l2(b2_out, b2_in))

    asm2_lp = WaveletAssembler2D(levels=2, wave='sym8', mode='tokens')
    z2 = asm2_lp(bands=b2_out)
    print("2D | pooled tokens shape:", z2.shape)

    # Build LP-only classifier and freeze HP path to avoid grad warnings
    inp2_lp = keras.Input((N, N, C))
    tb_lp = tok2_lp(inp2_lp)
    tb_lp = blk2_lp(tb_lp)
    vec_lp = asm2_lp(bands=tb_lp)
    logits2_lp = keras.layers.Dense(10)(vec_lp)
    model2_lp = keras.Model(inp2_lp, logits2_lp, name="wavelet_vit_2d_tokens_lp")

    tok2_lp.proj_hp.trainable = False
    try: blk2_lp.alpha_hp.assign(0.0)
    except Exception: pass
    blk2_lp.alpha_hp.trainable = False
    blk2_lp.hp_attn.trainable = False
    for m in blk2_lp.mixers: m.trainable = False

    model2_lp.compile(optimizer="adam", loss="mse", jit_compile=False)
    model2_lp.summary()
    model2_lp.fit(tf.random.normal((2, N, N, C)),
                  tf.random.normal((2, 10)), epochs=1, verbose=1)

    # Serialization round-trip (clone) — Keras 3 style
    try:
        model2_lp_clone = keras.models.clone_model(model2_lp)
    except TypeError:
        cfg = keras.saving.serialize_keras_object(model2_lp)
        model2_lp_clone = keras.saving.deserialize_keras_object(cfg)
    model2_lp_clone.set_weights(model2_lp.get_weights())
    model2_lp_clone.compile(optimizer="adam", loss="mse", jit_compile=False)
    model2_lp_clone.fit(tf.random.normal((2, N, N, C)),
                        tf.random.normal((2, 10)), epochs=1, verbose=1)

    # ========================= 2D (tokens) LP+HP tokens (HP gets gradients) =========================
    print("\n=== 2D (tokens, LP+HP pooled) ===")
    tok2_all = WaveletTokenizer2D(levels=2, wave='sym8', embed_dim=64)
    blk2_all = WaveletViTBlock2D(heads=2, key_dim=16, strength=0.25)
    asm2_all = WaveletAssembler2D(levels=2, wave='sym8', mode='tokens')

    inp2_all = keras.Input((N, N, C))
    tb_all = tok2_all(inp2_all)
    tb_all = blk2_all(tb_all)
    lp_vec = asm2_all(bands=tb_all)
    gap2 = keras.layers.GlobalAveragePooling2D()
    hp_vecs = [gap2(h) for lvl in tb_all for h in lvl['hp']]
    feat_all = keras.layers.Concatenate()([lp_vec] + hp_vecs) if hp_vecs else lp_vec
    logits2_all = keras.layers.Dense(10)(feat_all)
    model2_all = keras.Model(inp2_all, logits2_all, name="wavelet_vit_2d_tokens_lphp")
    model2_all.compile(optimizer="adam", loss="mse", jit_compile=False)
    _ = _grad_report(model2_all,
                     tf.random.normal((2, N, N, C)),
                     tf.random.normal((2, 10)),
                     keys=["proj_hp", "alpha_hp", "hp_attn", "orientation_mix"])
    model2_all.fit(tf.random.normal((2, N, N, C)),
                   tf.random.normal((2, 10)), epochs=1, verbose=1)

    # ========================= 3D (feature) =========================
    if HAVE_3D:
        print("\n=== 3D (feature) ===")
        try:
            orig_policy = tf.keras.mixed_precision.global_policy().name
            tf.keras.mixed_precision.set_global_policy('float32')
            print("3D test: precision policy -> float32")
        except Exception:
            orig_policy = None

        B3, D3, C3 = 1, 32, 1
        x3 = tf.random.normal((B3, D3, D3, D3, C3))

        # --- PR-2: DWT->IDWT transform round-trip sanity check (3D) ---
        try:
            dwt3_rt = DWT3D(wave='haar', clean=True)
            idwt3_rt = IDWT3D(wave='haar', clean=True)
            x3_rt_in = tf.cast(x3, tf.float32)
            w3_rt = dwt3_rt(x3_rt_in)
            x3_rt = idwt3_rt(w3_rt)
            rel3_rt = float(_rel_l2(x3_rt, x3_rt_in))
            print("3D | PR-2 (transform round-trip) rel-L2:", rel3_rt)
        except Exception as e:
            print("3D | PR-2 check skipped:", e)

        tok3 = WaveletTokenizer3D(levels=2, wave='haar', embed_dim=32)
        b3_in = tok3(x3)
        blk3_pr = WaveletViTBlock3D(heads=2, key_dim=8, strength=0.0)
        b3_out_pr = blk3_pr(b3_in)
        print("3D | PR (strength=0) bands rel-L2:", _bands_rel_l2(b3_out_pr, b3_in))

        blk3 = WaveletViTBlock3D(heads=2, key_dim=12, strength=1.0, level_decay=1.0)
        b3_out = blk3(b3_in)
        band_deltas = _bandspace_max_abs_delta(b3_in, b3_out)
        print("3D | band-space max|Δ| per level (LP, HP):", band_deltas)

        asm3 = WaveletAssembler3D(levels=2, wave='haar', mode='feature')
        y3_in = asm3(bands=b3_in)
        y3    = asm3(bands=b3_out)
        print("3D | feature rel-L2 (strength=1):", float(_rel_l2(y3, y3_in)))

        inp3 = keras.Input((D3, D3, D3, C3))
        tb3 = tok3(inp3); tb3 = blk3(tb3)
        feat3 = asm3(bands=tb3)
        out3 = keras.layers.Conv3D(4, 1)(feat3)
        model3 = keras.Model(inp3, out3, name="wavelet_vit_3d_feature")
        model3.compile(optimizer="adam", loss="mse", jit_compile=False)
        _ = _grad_report(model3,
                         tf.random.normal((1, D3, D3, D3, C3)),
                         tf.random.normal((1, D3, D3, D3, 4)),
                         keys=["proj_hp", "alpha_hp", "hp_attn", "orientation_mix"])
        model3.summary()
        model3.fit(tf.random.normal((1, D3, D3, D3, C3)),
                   tf.random.normal((1, D3, D3, D3, 4)), epochs=1, verbose=1)

        if orig_policy:
            tf.keras.mixed_precision.set_global_policy(orig_policy)
            print(f"Precision policy restored -> {orig_policy}")

    # ========================= 1D (tokens) =========================
    if HAVE_1D:
        print("\n=== 1D (tokens) ===")
        B1, T1, C1 = 2, 256, 3
        x1 = tf.random.normal((B1, T1, C1))

        # --- PR-2: DWT->IDWT transform round-trip sanity check (1D) ---
        try:
            dwt1_rt = DWT1D(wave='db8', clean=True)
            idwt1_rt = IDWT1D(wave='db8', clean=True)
            x1_rt_in = tf.cast(x1, tf.float32)
            w1_rt = dwt1_rt(x1_rt_in)
            x1_rt = idwt1_rt(w1_rt)
            rel1_rt = float(_rel_l2(x1_rt, x1_rt_in))
            print("1D | PR-2 (transform round-trip) rel-L2:", rel1_rt)
        except Exception as e:
            print("1D | PR-2 check skipped:", e)

        tok1 = WaveletTokenizer1D(levels=3, wave='db8', embed_dim=32)
        b1_in = tok1(x1)
        blk1_pr = WaveletViTBlock1D(heads=2, key_dim=8, strength=0.0)
        b1_out_pr = blk1_pr(b1_in)
        print("1D | PR (strength=0) bands rel-L2:", _bands_rel_l2(b1_out_pr, b1_in))

        blk1 = WaveletViTBlock1D(heads=2, key_dim=8, strength=0.25)
        b1_out = blk1(b1_in)
        print("1D | default strength bands rel-L2:", _bands_rel_l2(b1_out, b1_in))

        asm1 = WaveletAssembler1D(levels=3, wave='db8', mode='tokens')
        z1 = asm1(bands=b1_out)
        print("1D | pooled tokens shape:", z1.shape)

        inp1 = keras.Input((T1, C1))
        tb1 = tok1(inp1); tb1 = blk1(tb1)
        tvec1 = asm1(bands=tb1)
        out1 = keras.layers.Dense(5)(tvec1)
        model1 = keras.Model(inp1, out1, name="wavelet_vit_1d_tokens")
        model1.compile(optimizer="adam", loss="mse", jit_compile=False)
        model1.summary()
        model1.fit(tf.random.normal((2, T1, C1)),
                   tf.random.normal((2, 5)), epochs=1, verbose=1)

    print("\nDone.")



from __future__ import annotations
from typing import Dict, Tuple, Optional
import numpy as np
import os as _os
# Hardcode CPU-only by hiding all CUDA devices for TensorFlow
_os.environ.setdefault("CUDA_VISIBLE_DEVICES", "-1")
import tensorflow as tf

from .models import build_model, load_weights_best_effort
from .tiling import sliding_window_predict


def _choose_device(use_gpu: bool) -> str:
    # Force CPU regardless of meta flag
    return "/CPU:0"


def run_skull_strip(x4d: np.ndarray,
                    ss_cfg: Dict,
                    patch_size: int,
                    overlap: int,
                    use_gpu: bool,
                    wave: Optional[str] = None,
                    model=None) -> Tuple[np.ndarray, np.ndarray]:
    """Run skull stripping and return (mask3d, x4d_masked)."""
    device = _choose_device(use_gpu)
    with tf.device(device):
        n_classes = int(ss_cfg.get("n_classes", 2))
        loss = ss_cfg.get("loss", "focal")
        ckpt = ss_cfg["checkpoint"]
        model_type = ss_cfg.get("type", "uwavevitnet")
        class_weights = [0.1] * n_classes if "focal" in (loss or "").lower() else None
        if model is None:
            model = build_model(model_type, input_size=patch_size, n_classes=n_classes, loss=loss, class_weights=class_weights, wave=wave, compile_flag=True)
            load_weights_best_effort(model, ckpt)

        if patch_size == 256 and x4d.shape[:3] == (256, 256, 256):
            x_tf = tf.convert_to_tensor(x4d[None, ...])
            pred = model.predict(x_tf, verbose=0)[0]
        else:
            pred = sliding_window_predict(model, x4d, patch_size, overlap, n_classes, batch_size=1)

        labels = np.argmax(pred, axis=-1).astype(np.int16)
        mask = (labels > 0).astype(np.float32)
        x_masked = x4d.copy()
        x_masked[..., 0] *= mask
        return mask.astype(np.uint8), x_masked


def run_segmentation(x4d_masked: np.ndarray,
                     seg_cfg: Dict,
                     patch_size: int,
                     overlap: int,
                     use_gpu: bool,
                     wave: Optional[str] = None,
                     model=None) -> Tuple[np.ndarray, np.ndarray]:
    """Run GM/WM/CSF segmentation and return (probs4d, labels3d)."""
    device = _choose_device(use_gpu)
    with tf.device(device):
        n_classes = int(seg_cfg.get("n_classes", 4))
        loss = seg_cfg.get("loss", "focal")
        ckpt = seg_cfg["checkpoint"]
        model_type = seg_cfg.get("type", "ufdstvitnet")
        class_weights = [0.1] * n_classes if "focal" in (loss or "").lower() else None
        if model is None:
            model = build_model(model_type, input_size=patch_size, n_classes=n_classes, loss=loss, class_weights=class_weights, wave=wave, compile_flag=True)
            load_weights_best_effort(model, ckpt)

        if patch_size == 256 and x4d_masked.shape[:3] == (256, 256, 256):
            x_tf = tf.convert_to_tensor(x4d_masked[None, ...])
            probs = model.predict(x_tf, verbose=0)[0]
        else:
            probs = sliding_window_predict(model, x4d_masked, patch_size, overlap, n_classes, batch_size=1)

        labels = np.argmax(probs, axis=-1).astype(np.int16)
        return probs, labels
